#include "DistributeMat.h"

extern struct opts Options ;

void DistributeMatrixMondriaan(struct sparsematrix *pT, int P, double eps) {

    /* This function partitions the nonzeros of the matrix T
       into P parts, allowing a load imbalance of a fraction eps.

       Input: T sparse matrix,
              P number of parts, P >= 1,
              eps load imbalance fraction, eps >= 0.
       Output: T distributed sparse matrix.
               The nonzeros of part i are in positions
                   pT->Pstart[i], pT->Pstart[i+1]-1.

       The function picks the largest part, splits it into two parts
       of nearly-equal weight, and repeats this until all k parts
       are small enough to satisfy the load balance criterion,
           weight[i] <= (1 + eps) * totweight / P for i=0, ..., k-1,
       or when P parts have been reached.

    */

    int *procs ;     /* procs[i] = number of processors assigned to part i */
    int procslo ;    /* number of processors for smaller part after split */
    int procshi ;    /*    same for largest part */

    long totweight ; /* total weight of whole matrix */
    long maxweight ; /* maximum allowed weight of a matrix part */
    long *weight ;   /* weight[i] = weight of part i */
    long weightlo ;  /* smallest upper bound for part weight */
    long weighthi ;  /* largest upper bound */
  
    int i, j, k, dir=ROW, done ;

    totweight = ComputeWeight(*pT, 0, pT->NrNzElts-1, NULL) ;
    maxweight = ((1 + eps) * totweight ) / P ;  /* rounded down */  
  
    if ( Options.SplitStrategy == OneDimRow )
        dir = ROW ;
  
    if ( Options.SplitStrategy == OneDimCol )
        dir = COL ;
  
    if ( Options.SplitStrategy == FineGrain )
        dir = FINEGRAIN ;
  
    if ( Options.SplitStrategy == Alternate &&
        Options.Alternate_FirstDirection == Row )
        dir = ROW ;

    if ( Options.SplitStrategy == Alternate &&
        Options.Alternate_FirstDirection == Col )
        dir = COL ;
  
    if ( Options.SplitStrategy == Alternate &&
        Options.Alternate_FirstDirection == Ratio ) {  
        if ( pT->m > pT->n )
            dir = ROW;
        else if ( pT->n > pT->m )
            dir = COL;
        else if (Random1(0,1) == 0) /* random tie-breaking */
            dir = ROW;
        else
            dir = COL;
    }

    procs = (int *) malloc(P * sizeof(int)) ;
    weight = (long *) malloc(P * sizeof(long)) ;
    if ( procs == NULL || weight == NULL )
        Abort("DistributeMatrixMondriaan(): Not enough memory") ;
  
    /**** Split until all parts are small enough ****/  
    done = FALSE ;
    k = 1 ; /* k = current number of parts */ 
    procs[0] = P;
    weight[0] = totweight ;

    while (done == FALSE && k < P) {

        /* If desired, the number of processors assigned to each
           part is adjusted to reflect the actual outcome of
           previous splittings. This may allow for a larger
           imbalance parameter in future splits, aiming
           at lower communication. There is a trade-off:
               Load imbalance <--> communication volume. */
        if ( Options.LoadbalanceAdjust == AdjustYes ) 
            BalanceParts (weight, maxweight, P, k, procs) ;

        DetermineSplit (weight, maxweight, k, procs,
                        &i, &weightlo, &weighthi) ;

        if (i == -1) { 
            /* no more need to split  */
            done = TRUE ;
            break ;
        }
        
        /* Part i will be split into i, i+1 and
           the old parts i+1,..,k-1 will be shifted into i+2,..,k.
           Pstart will be  adjusted accordingly (within SplitMatrix) */
#ifdef INFO2
        printf("  ******** Split part %d from %d parts******** \n", i, k) ;
#endif
        if ( Options.SplitMethod == Simple ) 
            /* Simple split of the matrix only based on load balance,
               not minimising communication volume. 
               Useful for testing and debugging */
            SplitMatrixSimple(pT, k, i, weightlo, weighthi) ;
        else if ( Options.SplitMethod == KLFM )
            SplitMatrixKLFM(pT, k, i, dir, weightlo, weighthi) ;
        else
            Abort("DistributeMatrixMondriaan(): Unknown SplitMethod") ;
#ifdef INFO2
        printf("  Pstart[%d] = %ld ", i,  pT->Pstart[i]) ;
        printf("Pstart[%d] = %ld ", i+1,  pT->Pstart[i+1]) ;
        printf("Pstart[%d] = %ld ", i+2,  pT->Pstart[i+2]) ;
        printf("\n") ;
#endif

        /* Shift weight and procs */
        for ( j = k ; j > i+1 ; j-- ) {
            weight[j] = weight[j-1] ;
            procs[j] = procs[j-1] ;
        }

        k++ ; /* new number of parts */

        weight[i] = ComputeWeight(*pT, pT->Pstart[i], pT->Pstart[i+1]-1,NULL);
        weight[i+1] = ComputeWeight(*pT, pT->Pstart[i+1], pT->Pstart[i+2]-1,
                                    NULL);
        procslo = procs[i]/2 ;
        procshi = (procs[i]%2==0 ? procslo : procslo+1) ;
        
        if ( weight[i] <= weight[i+1] ) {
            procs[i] = procslo ;
            procs[i+1] = procshi ;
        } else { 
            procs[i] = procshi ;
            procs[i+1] = procslo ;
        }


        /* Check if there is a part that is too large */
        done = TRUE ;
        for ( j = 0 ; j < k ; j++ )
            if (weight[j] > maxweight) {
                done = FALSE ;
                break ;
            }
  
        /* Alternate between split directions, each time after
           the number of parts reaches a power of two.
           This strategy resembles the strategy in Mondriaan version 1 
           for powers of two, but it is not identical.
           (The part to be cut is the largest, and not the one
           determined by the numbering as in version 1.) */
        if (Options.SplitStrategy == Alternate && logb2(k+1) != logb2(k))
            dir = ! dir ;
    }

    /* Check if each part is small enough */
    for (j = 0 ; j < k ; j++)
        if (weight[j] > maxweight) 
#ifdef INFO0
            printf("WARNING: part %d too large \n",j) ;
#endif

    /* Set Pstart for empty parts  */
    for (j = k+1 ; j <= P ; j++)
        pT->Pstart[j] = pT->Pstart[k] ;

#ifdef INFO2
    printf("  Number of parts = %d \n", k) ;
    printf("  Pstart = ") ;
    for (j = 0 ; j <= P ; j++)
        printf("%ld ", pT->Pstart[j]) ;
    printf("\n\n") ;
#endif      

    /* Set matrix type code to distributed */
    pT->MMTypeCode[0] = 'D' ;

    free(weight) ;
    free(procs) ;

    return ;

} /* end DistributeMatrixMondriaan */  


long ComputeWeight (struct sparsematrix A, long lo, long hi, long *wnz) {

    /* This function computes the weight of the matrix part containing
       the nonzeros in positions lo..hi. Each nonzero has weight 1,
       except the dummies which have weight 0. 

       In case of a symmetric, skew-symmetric, or hermitian matrix
       and a single entry representing a pair of nonzeros,
       the off-diagonal nonzeros have weight 2.

       If wnz <> NULL, the weights of the individual nonzeros are also
       returned: wnz[t] = weight of nonzero lo+t, for 0 <= t <= hi-lo. 

       In case of a column-weighted matrix, the returned weight is the total
       weight of the columns that have a nonzero in a position lo..hi.
       No individual weights are returned.

   */

    int *marked ;
    long t, j, weight, Ndum, Ndiag ;
  
    if (lo > hi) 
        return 0 ;
       
    if ( A.MMTypeCode[0] == 'W' && A.NrColWeights > 0 ) {
        /* The weight is determined by the weights 
           of columns containing nonzeros and not by the nonzeros */
        
        /* Mark processed columns to avoid counting their weight twice */
        marked = (int *) malloc(A.NrColWeights * sizeof(int)) ;
        if ( marked == NULL )
            Abort("ComputeWeight(): Not enough memory") ;
        for (j = 0; j < A.NrColWeights; j++)
            marked[j] = FALSE ;
    
        weight = 0 ;
        for ( t = lo ; t <= hi ; t++ ) {
            j = A.j[t] ;
            if (marked[j]==FALSE) {
                weight +=  A.ColWeights[j] ; 
                marked[j] = TRUE ;
            }
        }
        
        free(marked) ;
        
    } else {
                        
        weight = hi - lo + 1 ;
    
        /* Initialise individual weights to standard value */
        if (wnz != NULL)
            for ( t = lo ; t <= hi ; t++ )
                wnz[t-lo] = 1;

        /* Count the dummies and set their individual weights to 0 */
        Ndum = 0 ;
        if ( A.m == A.n && A.NrDummies > 0) { 
            for ( t = lo ; t <= hi ; t++ )
                if (A.i[t] == A.j[t] && A.dummy[A.i[t]] ) {
                      /* nonzero t is a dummy */
                      Ndum++ ;
                      if (wnz != NULL) 
                          wnz[t-lo] = 0; 
                }
            weight -= Ndum ; /* weight = number of nondummy nonzeros */
        }
  
        /* Count the number of diagonal entries and set the individual
           weights of off-diagonal entries to 2 */
        if ( A.m == A.n &&
             (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' ||
              A.MMTypeCode[3]=='H') &&
             Options.SymmetricMatrix_UseSingleEntry == SingleEntYes ) {
            Ndiag = 0;
            for ( t = lo ; t <= hi ; t++ ) {
                if (A.i[t] == A.j[t]) 
                    Ndiag++ ; /* individual weight is already OK */
                else if (wnz != NULL )
                    wnz[t-lo] = 2;   
            }

            /* Double the total weight, except for the Ndiag - Ndum
               nondummy nonzeros on the diagonal */
            weight = 2*weight - (Ndiag - Ndum) ; 
        }
    }

    return weight ;
  
} /* end ComputeWeight */


void BalanceParts (long *weight, long maxweight, int P, int k, int *procs) {
 
    /* This function balances the parts of a partitioning by assigning
       a number of processors to each part in proportion to its weight. 
       If a part has weight <= maxweight, its number of processors will be set
       to -1 meaning that it need not be split anymore.

       Input: weight[i] = weight of part i, weight[i] >= 0, 0 <= i < k,
              maxweight = maximum allowed weight per processor
                          after partitioning, maxweight >= 0,
              P = total number of processors available,
              k = number of parts, 1 <= k <= P. 
       Output: procs[i] = number of processors assigned to part i. */
 
    long totweight, *surplus, *I;
    int i, j, nfinished, nextra, p, q ;
  
    if (k <= 0)
        return ; /* no need for balancing */
    else if (k == 1) {
        if (weight[0] <= maxweight)
            procs[0] = -1;
        else
            procs[0] = P;
    } else {
        /* Assign finished parts and compute total remaining weight */
        totweight = 0 ;
        nfinished = 0; /* number of finished parts */
        for(i = 0 ; i < k ; i++) {
            if (weight[i] <= maxweight) { 
                procs[i] = -1;
                nfinished++;
            } else {     /* weight[i] > 0 */
                procs[i] = 0; /* to be set later */
                totweight += weight[i] ;
            }
        }
        p = P - nfinished ; /* number of processors still available,
                               i.e., not assigned to finished parts */
                
        surplus = (long *)malloc(k*sizeof(long)); 
        if( surplus == NULL )
            Abort("BalanceParts(): Not enough memory");
                
        q = 0 ;
        for(i=0 ; i < k ; i++) {
            if (procs[i] == 0 && totweight > 0) {
                /* use long longs for intermediate results to prevent overflow */
                procs[i] = ((long long) weight[i] * (long long)p) / totweight ;
                           /* = weight[i] / avgweight */

                surplus[i] = (long long)weight[i] * (long long)p -
                             (long long)procs[i] * (long long)totweight ;
                             /* = (weight[i] - procs[i]*avgweight) * p.
                             The scaling by a factor p is to obtain integers. */

                q += procs[i] ; /* total number of processors assigned so far
                                   for unfinished parts */
            } else 
                surplus[i] = 0;
        }
        
        I = QSort(surplus, (long)k) ;
        nextra = 0;
        for(j=0 ; j < k ; j++) {
            i = I[j] ;
            if (procs[i] >= 0 && nextra < p - q) {
                /* Assign an extra processor to this part because of a large surplus.
                   The number of extra processors available is p-q.  */
                procs[i]++ ;
                nextra++ ;
            }
        }
       
        free(I) ;
        free(surplus) ;

    }
    return ;
    
} /* end BalanceParts */


void DetermineSplit (long *weight, long maxweight, int k, int *procs, 
                     int *isplit, long *weightlo, long *weighthi) {

    /* This function determines the part to split in the next bipartitioning
       and computes upper bounds for the weight of the smallest and largest
       parts resulting from the split. 
       
       The weight bounds are based on a split into equal parts if procs[isplit]
       is even, and nearly equal parts if it is odd. All splits of part isplit
       that remain to be done allow the weight imbalance to grow with a constant
       factor, or, alternatively, by an increasing or decreasing factor.
    .  The factor is chosen such that all parts resulting from isplit
       at the end of the partitioning have weight <= maxweight.
       
       Input: weight[i] = weight of part i, weight[i] >= 0, 0 <= i < k,
              maxweight = maximum allowed weight per processor
                          after partitioning, maxweight >= 0,
              k = number of parts,
              procs[i] = number of processors assigned to part i,
                         or -1 if part need not be split anymore.
              
       Output: isplit = index of part to be split,
                        or -1 if no need to split any more.
               weightlo = smallest upper bound for part weight,
               weighthi = largest upper bound for part weight.
               procs is modified: some finished parts are marked by -1
                     and the rest must be reinitialised afterwards. */
       
       int plo, phi, qlo, qhi, i, ilarge ;
       long wlarge ;
       double wlo, whi, eps, deltalo, deltahi ;
       
       if (k <= 0) {
           *isplit = -1 ;
           return ; /* no need for splitting */
       }
   
       /* Find part of largest weight */
       wlarge = -1;
       ilarge = -1;
       for (i = 0 ; i < k ; i++) {
           if (weight[i] > wlarge && procs[i] >= 0 ) {
                wlarge = weight[i] ;
                ilarge = i ;
           }
       }
       i = ilarge ;

       if (wlarge == -1)
           *isplit = -1;
       else {
           *isplit = i ;
           if (weight[i] > procs[i]*maxweight ) {
               /* Desired split is infeasible, so grab an extra processor,
                  which is almost always available. */
               procs[i]++ ;
#ifdef INFO2
               printf("    Warning: extra processor grabbed for part %d\n",i) ;
#endif
           }
           if (weight[i] > procs[i]*maxweight )
               /* Desired split is still infeasible */
               Abort("DetermineSplit(): no feasible split possible\n") ;

           if (procs[i]==1) {
               /* No split needed. Part is feasible, so mark it as finished. */
               procs[i] = -1 ;
               return ;
           }

           plo = procs[i]/2 ;
           phi = (procs[i]%2==0 ? plo : plo+1) ;

           /* maximum number of splits (including current split)
              needed to reach leaf of splitting tree */
           qlo = logb2(plo) + 1; 
           qhi = logb2(phi) + 1;
      
           eps =  (maxweight*procs[i])/(double)weight[i] - 1.0 ; /* eps >= 0 */

           deltalo = eps ; /* default value for plo = 1 */
           deltahi = eps ;
           
           /* Splitting imbalance delta(0) for part i is determined by inequality
                   (1+delta(0))...(1+delta(q-1)) weight[i]/procs[i] <= maxweight,
              i.e. (1+delta(0))...(1+delta(q-1)) <= 1 + eps. */
    
           if ( Options.LoadbalanceStrategy == Decrease ) {
               /* take delta(0) large */
               if ( plo > 1 )
                   deltalo = eps/2 ;
               if ( phi > 1 )
                   deltahi = eps/2 ;
           } else if ( Options.LoadbalanceStrategy == Increase ) {
               /* take delta(0) small */ 
               if ( plo > 1 )
                   deltalo = eps/(2*plo) ;
               if ( phi > 1 )
                   deltahi = eps/(2*phi) ;
           } else if ( Options.LoadbalanceStrategy == Constant ) {
               /* take delta(0) assuming constant growth */
               if ( plo > 1 )
                   deltalo = eps/qlo ;
               if ( phi > 1 )
                   deltahi = eps/qhi ;
           } else
               Abort("DetermineSplit(): Unknown LoadbalanceStrategy") ;

           wlo = (weight[i] / (double)procs[i]) * plo ;
           whi = (weight[i] / (double)procs[i]) * phi ;
           *weightlo = (long) ( (1.0 + deltalo) * wlo ) ;
           *weighthi = (long) ( (1.0 + deltahi) * whi ) ;
        }
              

 } /* end DetermineSplit */


int logb2 (int n) {

    /* This function computes the base-2 log of n, n >= 1,
       rounded up to the nearest integer. */

    int k, i;
    char roundup ;

    k = 0 ;
    i = n; /* n = (b(m-1) ...b(0)) in binary notation
                                   with b(0) the least significant bit
                                   and b(m-1) = 1 the most significant bit */
    roundup = FALSE ;

    while (i >= 2) {
        /* i = (b(m-1) ...b(k)) */
        if (i%2 == 1)
            roundup = TRUE;
        i /= 2;
        k++;
    }
    /* i = 1 = (b(m-1)),
       k = number of divisions performed = m-1*/

    if (roundup)
        k++ ;

    return k ;

} /* end logb2 */


void SplitMatrixKLFM(struct sparsematrix *pT, int k, int i, int dir,
                     long weightlo, long weighthi) {
    
    /* This function splits part i of the sparse matrix T into two parts,
       the first with weight <= weightlo and the second with weight <= weighthi.
       The split is performed by translating the matrix into a hypergraph
       and bipartitioning this hypergraph, trying to keep vertices
       from the same hyperedge together as much as possible.
       The bipartitioning employs a multilevel hypergraph partitioner
       based on the Kernighan-Lin/Fiduccia-Mattheyses algorithm.

       Input: T sparse matrix,
              k current number of parts, 1 <= k < P,
              i number of part to be split, 0 <= i < k,
              dir = direction of split. The input value is used,
                    except for the LocalBest, Hybrid strategies,
                    where the value is overruled by this function,
                    and for LocalRatio, where it is set elsewhere.
              weightlo = smallest upper bound for part weight, belongs to part 0
              weighthi = largest upper bound for part weight, belongs to part 1.
              
       Output: T sparse matrix.
               The nonzeros of the new part i (the first part) are in positions
                   pT->Pstart[i], pT->Pstart[i+1]-1.
               The nonzeros of the new part i+1 (the second) are in positions
                   pT->Pstart[i+1], pT->Pstart[i+2]-1.
               All parts > i+1 have been shifted.
       
    */

    long lo, hi, mid, nz, weight;
    int j, bestdir=ROW, bestcomm ;
    struct sparsematrix A ;
    struct biparthypergraph HG, HGRow, HGCol, HGFine ;
    
    lo = pT->Pstart[i] ;
    hi = pT->Pstart[i+1]-1 ;
    
    /* Shift Pstart for parts > i */
    for (j = k ; j > i ; j--)
        pT->Pstart[j+1] = pT->Pstart[j] ;
        
    nz = hi-lo+1 ;
    if ( nz == 0 ) 
        return ; /* Pstart[i] = Pstart[i+1] */
    
    weight = ComputeWeight(*pT, lo, hi, NULL) ;
    if (weight > weightlo + weighthi )
        /* Desired split is infeasible */
        Abort("SplitMatrixKLFM(): desired hypergraph split is infeasible\n") ;
     
    /* Copy info from T to A */ 
    A = *pT ;            /* A has same size as T, and same other parameters, */
    A.i = &(pT->i[lo]) ; /* but only a subset of the nonzeros */
    A.j = &(pT->j[lo]) ;
    if (A.MMTypeCode[2] != 'P')
        A.ReValue = &(pT->ReValue[lo]) ;
    if (A.MMTypeCode[2] == 'C')
        A.ImValue = &(pT->ImValue[lo]) ;
    A.NrNzElts = nz ;
  
    if ( Options.SplitStrategy == LocalBest ||
         Options.SplitStrategy == Hybrid ) {

#ifdef INFO2
        printf("**** Try split in row direction \n") ;
#endif    
        HGRow = SparseMatrix2BiPartHyperGraph(A, ROW) ;
        RunMLGraphPart(&HGRow, weightlo, weighthi) ;
  
#ifdef INFO2
        printf("**** Try split in col direction \n") ;
#endif   
        HGCol = SparseMatrix2BiPartHyperGraph(A, COL) ;
        RunMLGraphPart(&HGCol, weightlo, weighthi) ;
  
#ifdef INFO2
        printf("Communication: row=%ld col=%ld\n", 
               HGRow.OptComm, HGCol.OptComm) ;
#endif
        
        if ( HGRow.OptComm < HGCol.OptComm )
            bestdir = ROW ;
        else if ( HGCol.OptComm < HGRow.OptComm )
            bestdir = COL ;
        else if (Random1(0,1) == 0) /* no preference, random tie-breaking */
            bestdir = ROW ;
        else
            bestdir = COL;
        bestcomm = MIN(HGRow.OptComm, HGCol.OptComm) ;

        if ( Options.SplitStrategy == Hybrid ) {

#ifdef INFO2
            printf("**** Try split in finegrain direction \n") ;
#endif
            HGFine = SparseMatrix2BiPartHyperGraph(A, FINEGRAIN) ;
            RunMLGraphPart(&HGFine, weightlo, weighthi) ;

#ifdef INFO2
            printf("Communication: finegrain=%ld\n", HGFine.OptComm ) ;
#endif
            if ( HGFine.OptComm < bestcomm )
                bestdir = FINEGRAIN ;
            /* in case of equality, we prefer the row or column direction */
        }
  
#ifdef INFO2
        if ( bestdir == ROW )
            printf("**** Row direction chosen\n") ;
        else if ( bestdir == COL )
            printf("**** Col direction chosen\n") ;
        else
            printf("**** Finegrain direction chosen\n") ;
#endif   
  
        if ( bestdir == ROW )
            BiPartHyperGraph2SparseMatrix(HGRow, *pT, lo, hi, &mid) ;
        else if ( bestdir == COL )
            BiPartHyperGraph2SparseMatrix(HGCol, *pT, lo, hi, &mid) ;
        else
            BiPartHyperGraph2SparseMatrix(HGFine, *pT, lo, hi, &mid) ;
  
        if ( Options.SplitStrategy == Hybrid ) 
            DeleteBiPartHyperGraph(&HGFine) ;
        DeleteBiPartHyperGraph(&HGCol) ;
        DeleteBiPartHyperGraph(&HGRow) ;
        
    } else {

#ifdef INFO2
        if ( Options.SplitStrategy != LocalRatio ) {
            if ( dir == ROW )
                printf("**** Split in row direction \n") ;
            else
                printf("**** Split in col direction \n") ;
        } /* For LocalRatio, dir will be set inside
             SparseMatrix2BiPartHyperGraph */
#endif
  
        HG = SparseMatrix2BiPartHyperGraph(A, dir) ;
        RunMLGraphPart(&HG, weightlo, weighthi) ;
        BiPartHyperGraph2SparseMatrix(HG, *pT, lo, hi, &mid) ;
        DeleteBiPartHyperGraph(&HG) ;
    }

    /* register new splitting point */ 
    pT->Pstart[i+1] = mid ;

} /* end SplitMatrixKLFM */


void SplitMatrixSimple(struct sparsematrix *pT, int k, int i,
                       long weightlo, long weighthi) {

    /* This function splits part i of the sparse matrix T into two parts,
       the first with weight <= weightlo and the second with weight <= weighthi.
       The split is performed by simply cutting somewhere in the middle
       of the array of nonzeros. Nonzeros are not moved.
       Input: T sparse matrix,
              k current number of parts, 1 <= k < P,
              i number of part to be split, 0 <= i < k,
              weightlo = smallest upper bound for part weight, belongs to part 0,
              weighthi = largest upper bound for part weight, belongs to part 1.
              
       Output: T sparse matrix.
               The nonzeros of the new part i (the first part) are in positions
                   pT->Pstart[i], pT->Pstart[i+1]-1.
               The nonzeros of the new part i+1 (the second) are in positions
                   pT->Pstart[i+1], pT->Pstart[i+2]-1.
               All parts > i+1 have been shifted.
               
       NB: This function works only if the weights are based on the number
       of nonzeros.  Dummy nonzeros with  weight 0 are allowed, and so are
       nonzeros with weight 2 caused by a symmetric matrix with single-entry
       storage. The function does not work, however, for weights based on
       matrix columns.
       
    */
       
    long lo, hi, nz, weight, surplus, ideal, wtot, t, *wnz ;
    int j ;
    
    if ( pT->MMTypeCode[0] == 'W' && pT->NrColWeights > 0 ) 
        Abort("SplitMatrixSimple(): cannot be used with column-weighted matrix") ;

    lo = pT->Pstart[i] ;
    hi = pT->Pstart[i+1]-1 ;
    
    /* Shift Pstart for parts > i */
    for (j = k ; j > i ; j--)
        pT->Pstart[j+1] = pT->Pstart[j] ;
        
    nz = hi-lo+1 ;
    if ( nz == 0 ) 
        return ; /* Pstart[i] = Pstart[i+1] */
        
    /* Allocate memory for storing weights 0, 1, 2 of individual nonzeros */
    wnz = (long *) malloc(nz * sizeof(long)) ; 
    if ( wnz == NULL )
        Abort("SplitMatrixSimple(): Not enough memory") ;
 
    weight = ComputeWeight(*pT, lo, hi, wnz) ;
       
    if (weight > weightlo + weighthi )
        /* Desired split is infeasible */
        Abort("SplitMatrixSimple(): desired simple split is infeasible\n") ;
        
    /* The surplus is equally divided over the two parts */
    surplus = weightlo + weighthi - weight;
    ideal = MAX(weightlo - surplus/2,0) ; /* ideal weight of new part i */
    
    /* Find first legal split */
    wtot = 0 ;
    t = 0 ;
    while (t < nz && wtot + weighthi < weight ) {
        wtot += wnz[t] ;
        t++ ; /* t = number of nonzeros included in smallest part */
    }  
    if (wtot > weightlo) 
        /* Desired simple split is infeasible, because a nonzero must be split.
           This is only likely to occur if bounds are too tight, and 
           then only if weights of 2 are used.*/
        Abort("SplitMatrixSimple(): simple split causes split of a nonzero\n") ;
         
    /* If possible, improve until ideal split */
    while (t < nz && wtot + wnz[t] <= ideal) {
        wtot += wnz[t] ;
        t++ ;
    }  
     
    pT->Pstart[i+1] = pT->Pstart[i] + t ;
    
    free(wnz) ;
    return ;
    
} /* end SplitMatrixSimple */
