#include "DistributeVecOrigEq.h"

extern struct opts Options ;

long DistributeVecOrigEq(struct sparsematrix A, int *U, int *V) {

    /*  This function determines a vector distribution U = V by the
        vector partitioning algorithm for equal input and output distribution
        in the original SIAM Review paper [1],
        which tries to minimise the communication cost, defined as
        the maximum over all processors of max(sends,recvs).

        The function has not been fully optimised, since in general
        there will not be much choice in assigning the vector components.

        [1] B. Vastenhouw and R. H. Bisseling,
            A two-dimensional data distribution method
            for parallel sparse matrix-vector multiplication",
            SIAM Review, Vol. 47, No. 1 (2005) pp. 67-95.

        Input:  initialised A,
                uninitialised arrays U, V of length A.m = A.n,
                representing output and input distributions for u = Av.

        Output: vector distributions U, V, with U=V, 
                which are arrays of length A.m
                with values between 0 and P-1, where P= A.NrProcs.
                The function returns maxcom = max(sends,recvs). */
 

    int P, Pact, q, r, proc ;
    long l, j, total, LB, actbound, bestbound, 
         ComVol, MaxOut, MaxIn, MaxCompnts, TotCompnts, maxcom,
         *ProcHistogramU, *ProcHistogramV , *ProcHistogramUV ;
    
    /* Arrays of length A.n for vector v */
    int *NprocsV;   /* NprocsV[j] is the number of processors
                       that occur in column j */
 
    /* Communication matrix for vector v, stored in compressed column storage (CCS).
       Includes all columns of the original matrix. */
    int *procindexV; /* array of length at most V+A.n, where V is the
                       communication volume for the vector v.
                       procindexV[procstartV[j]..procstartV[j+1]-1]
                       contains the processor numbers of the processors
                       that occur in column j */
    long *procstartV; /* array of length A.n+1 containing the starts of the
                       columns of the communication matrix */

    /* Arrays of length P for vector v */
    long *NsV;       /* NsV[q] = the current number of vector components
                       sent by processor q to other processors */
    long *NrV;       /* NrV[q] = the current number of vector components
                       received by processor q from other processors */
    long *NvV;       /* NvV[q] = the current number of vector components
                       owned by processor q */
    long *SumsV;     /* SumsV[q] = the current number of communications
                       (sends or receives) of processor q, taking inevitable
                       communications into account from the start */
    
    /* The same for vector u */
                       
    int *NprocsU;   /* NprocsU[j] is the number of processors
                       that occur in row j */
    int *procindexU; /* procindexU[procstartU[j]..procstartU[j+1]-1]
                       contains the processor numbers of the processors
                       that occur in row j */
    long *procstartU; 
    long *NsU, *NrU, *NvU, *SumsU;     
        
    int *NprocsUV;  /* NprocsUV[j] is the number of processors that occur 
                       both in row j and in column j */
  
    /*## Check that the matrix is square: ##*/
    if ( A.m != A.n )
        Abort("DistributeVecOrigEq(): U and V cannot be distributed the same way"
              "because the matrix is not square") ;
  
    l = A.n ;
    P = A.NrProcs ;
 
    /*## Allocate processor arrays: ##*/
    ProcHistogramU = (long *) malloc((P+1) * sizeof(long)) ;
    ProcHistogramV = (long *) malloc((P+1) * sizeof(long)) ;
    ProcHistogramUV = (long *) malloc((P+1) * sizeof(long)) ;
    NsU = (long *) malloc(P * sizeof(long)) ;
    NsV = (long *) malloc(P * sizeof(long)) ;
    NrU = (long *) malloc(P * sizeof(long)) ;
    NrV = (long *) malloc(P * sizeof(long)) ;
    NvU = (long *) malloc(P * sizeof(long)) ;
    NvV = (long *) malloc(P * sizeof(long)) ;
    SumsU = (long *) malloc(P * sizeof(long)) ;
    SumsV = (long *) malloc(P * sizeof(long)) ;
    if ( ProcHistogramU == NULL || ProcHistogramV == NULL || 
         ProcHistogramUV == NULL ||  
         NsU == NULL || NrU == NULL || NvU == NULL || SumsU == NULL ||
         NsV == NULL || NrV == NULL || NvV == NULL || SumsV == NULL )
        Abort("DistributeVecOrigEq(): Not enough memory for processor arrays") ;
        
    /*## Clear memory: ##*/
    for ( q = 0 ; q < P ; q++ ) {
        NsU[q] = NrU[q] = NvU[q] = 0 ;
        NsV[q] = NrV[q] = NvV[q] = 0 ;
    }
  
    /*## STEP 1: 
      ##  Generate communication matrices, histograms and
          initialise intersections and sums: ##*/
      
    /* Initialise CRS data structure for communication matrix of U */
    NprocsU = (int *)malloc(l*sizeof(int));
    procstartU = (long *)malloc((l+1)*sizeof(long));
    if(NprocsU == NULL || procstartU == NULL)
        Abort("DistributeVecOrigEq(): Not enough memory");

    InitNprocs(A, COL, NprocsU);

    total = 0;
    for(j=0 ; j<l ; j++)
        total += NprocsU[j] ;
    procindexU = (int *)malloc(total*sizeof(int));
    if(procindexU == NULL )
        Abort("DistributeVecOrigEq(): Not enough memory");

    InitProcindex(A, COL, NprocsU, procstartU, procindexU);
    
    /* Initialise CCS data structure for communication matrix of V */
    NprocsV = (int *)malloc(l*sizeof(int));
    procstartV = (long *)malloc((l+1)*sizeof(long));
    if(NprocsV == NULL || procstartV == NULL)
        Abort("DistributeVecOrigEq(): Not enough memory");

    InitNprocs(A, ROW, NprocsV);

    total = 0;
    for(j=0 ; j<l ; j++)
        total += NprocsV[j] ;
    procindexV = (int *)malloc(total*sizeof(int));
    if(procindexV == NULL )
        Abort("DistributeVecOrigEq(): Not enough memory");

    InitProcindex(A, ROW, NprocsV, procstartV, procindexV);

    NprocsUV = (int *) malloc(l * sizeof(int)) ;
    if ( NprocsUV == NULL )
        Abort("DistributeVecOrigEq(): Not enough memory") ;
    InitIntersection(l, P, procstartU, procindexU, procstartV, procindexV,
                     NprocsUV) ;
    
    GenerateHistogram(NprocsU, l, 0, P, ProcHistogramU) ;
    GenerateHistogram(NprocsV, l, 0, P, ProcHistogramV) ;
    GenerateHistogram(NprocsUV, l, 0, P, ProcHistogramUV) ;
  
    /* Calculate and print communication statistics for vector v */ 
    CalcCom(A, NULL, ROW, &ComVol, &MaxOut, &MaxIn, &MaxCompnts, &TotCompnts) ;
    CalcLocalLowerBound(A, ROW, &LB, &Pact);
    if (Pact > 0)
        actbound= (ComVol%Pact == 0 ? ComVol/Pact : ComVol/Pact +1);
    else
        actbound= 0;
    bestbound= MAX(LB,actbound);

#ifdef INFO
    printf("\nNr matrix columns owned by p processors:\n" ) ;
    PrintHistogram(0, P, ProcHistogramV);
    printf("Communication for vector v (after matrix distribution):\n") ;
    printf("  vol           : %ld \n", ComVol) ;
    printf("  active procs  : %d \n", Pact) ;
    printf("  best bound    : %ld \n", bestbound) ;
#endif
#ifdef INFO2
    printf("    active bound: %ld \n", actbound) ;
    printf("    local bound : %ld \n", LB) ;
#endif 
  
    /* Calculate and print communication statistics for vector u */  
    CalcCom(A, NULL, COL, &ComVol, &MaxOut, &MaxIn, &MaxCompnts, &TotCompnts) ;
    CalcLocalLowerBound(A, COL, &LB, &Pact); 
    if (Pact > 0) 
        actbound= (ComVol%Pact == 0 ? ComVol/Pact : ComVol/Pact +1); 
    else 
        actbound= 0; 
    bestbound= MAX(LB,actbound); 

#ifdef INFO
    printf("\nNr matrix rows owned by p processors:\n" ) ;
    PrintHistogram(0, P, ProcHistogramU); 
    printf("Communication for vector u (after matrix distribution):\n") ;
    printf("  vol           : %ld \n", ComVol) ; 
    printf("  active procs  : %d \n", Pact) ; 
    printf("  best bound    : %ld \n", bestbound) ;
#endif 
#ifdef INFO2 
    printf("    active bound: %ld \n", actbound) ; 
    printf("    local bound : %ld \n", LB) ; 
#endif 

/* Print statistics for intersection of vectors u and v */
#ifdef INFO 
    printf("\nNr rows/columns with p processors in the intersection\n");
    printf("                           (after matrix distribution):\n") ;
    PrintHistogram(0, P, ProcHistogramUV); 
    printf("\n");
#endif
  
    /* Initialise sums and vector distributions */
    InitSums(l, P, procstartU, procindexU, SumsU) ;
    InitSums(l, P, procstartV, procindexV, SumsV) ;
    
    /* All columns are unowned at the start */
    for(j=0 ; j<l ; j++) {
        U[j] = -1 ;
        V[j] = -1 ;
    }
  
    /*## STEP 2: 
      ##  Assign vector components with one processor in the intersection : ##*/ 
    for(j=0 ; j<l ; j++) {
        if ( NprocsUV[j] == 1 ) {
            q = procindexU[procstartU[j]] ;
  
            /* Assign U[j] and V[j] to processor q: */
            AssignColumnToProc(U, procstartU, procindexU, NsU, NrU, NvU, SumsU,
                               j, q);
            AssignColumnToProc(V, procstartV, procindexV, NsV, NrV, NvV, SumsV,
                               j, q);
        }
    }
  
    /*## STEP 3: 
      ##  Assign vector components with more than one processor in the 
      ##  intersection :##*/ 
    for(j=0 ; j<l ; j++) {
        if ( NprocsUV[j] > 1 ) {
      
            /* Find a processor with the lowest total sum: */
            q = FindProcLowestSum(procstartU, procindexU, j, NprocsUV[j],
                                  SumsU, SumsV ) ;
 
            /* If there is more than one processor having the minimum sum
               we could look at a secondary objective. 
               However, we do not use secondary objectives here. */
    
            /* Assign U[j] and V[j] to this processor: */
            AssignColumnToProc(U, procstartU, procindexU, NsU, NrU, NvU, SumsU, 
                               j, q) ;
            AssignColumnToProc(V, procstartV, procindexV, NsV, NrV, NvV, SumsV,
                               j, q) ;
        }
    }   
  
  
    /*## STEP 4: 
      ##  Assign vector components with no processors in the
          intersection of the processor lists of u and v: ##*/ 
  
    /*## Step 4A:
      ##  Assign vector components j with NprocsU[j]=0 XOR NprocsV[j]=0: ##*/
    for(j=0 ; j<l ; j++) {
        if ( NprocsUV[j] == 0 ) {

            /* Find a processor with the lowest total sum: */
            if ( NprocsU[j] != 0 && NprocsV[j] == 0 ) 
                q = FindProcLowestSum(procstartU, procindexU, j,NprocsU[j],
                                      SumsU,SumsV);
            else if ( NprocsU[j] == 0 && NprocsV[j] != 0 ) 
                q = FindProcLowestSum(procstartV, procindexV, j, NprocsV[j],
                                      SumsU,SumsV);
      
            /* Assign U[j] and V[j] to this processor: */
            AssignColumnToProc(U, procstartU, procindexU, NsU, NrU, NvU, SumsU,
                               j, q) ;
            AssignColumnToProc(V, procstartV, procindexV, NsV, NrV, NvV, SumsV,
                               j, q) ;
        }
    }
  
    
    /*## Step 4B:
      ##  Assign vector components j with NprocsU[j]!=0 and NprocsV[j]!=0: ##*/
    for(j=0 ; j<l ; j++) {
        if ( NprocsUV[j] == 0 && NprocsU[j] != 0 && NprocsV[j] != 0) {

            /* Find a processor with the lowest total sum: */
            q = FindProcLowestSum (procstartU, procindexU, j, NprocsU[j],
                                   SumsU,SumsV);
            r = FindProcLowestSum (procstartV, procindexV, j, NprocsV[j],
                                   SumsU,SumsV);   
    
            /* If the matrix is partitioned one-dimensionally,
               then the vector is partitioned conformally.
               Otherwise, the processor with the lowest total sum is chosen.
               Ties are decided by flipping a coin, to avoid favouring 
               the input or output vector. */
            if ( Options.SplitStrategy == OneDimRow )
               proc = q ;
            else if ( Options.SplitStrategy == OneDimCol )
               proc = r ;
            else if ( SumsU[q] + SumsV[q] < SumsU[r] + SumsV[r] )
               proc = q ;
            else if ( SumsU[q] + SumsV[q] > SumsU[r] + SumsV[r] )
               proc = r ;
            else if (Random1(0,1) == 0)
               proc = q ;
            else
               proc = r ;
  
            /* Assign U[j] and V[j] to this processor: */
            AssignColumnToProc(U, procstartU, procindexU, NsU, NrU, NvU, SumsU, 
                               j, proc) ;
            AssignColumnToProc(V, procstartV, procindexV, NsV, NrV, NvV, SumsV,
                               j, proc) ;
        }
    }
    
    
    /*## Step 4C:
      ##  Assign vector components j with NprocsU[j]=0 and NprocsV[j]=0: ##*/
    AssignRemainingColumns(l, P, U, NvU) ;
    for(j=0 ; j<l ; j++) {
        if (V[j] == -1)
            V[j] = U[j] ;
        else if (V[j] != U[j]) 
            Abort("DistributeVecOrigEq(): Internal error step 4C") ;
    }
  
    CalcCom(A, V, ROW, &ComVol, &MaxOut, &MaxIn, &MaxCompnts, &TotCompnts) ;
    maxcom =  MAX(MaxOut,MaxIn) ;
    
    /* Check whether total number of components equals length: */
    if ( TotCompnts != l )
        Abort("DistributeVecOrigEq(): Internal error: TotCompnts != length") ;
  
#ifdef INFO
    PrintCom(P, l, ROW, ComVol, MaxOut, MaxIn, MaxCompnts, TotCompnts) ;
#endif
    
    CalcCom(A, U, COL, &ComVol, &MaxOut, &MaxIn, &MaxCompnts, &TotCompnts) ;
    maxcom +=  MAX(MaxOut,MaxIn) ;

    /* Check whether total number of components equals length: */
    if ( TotCompnts != l )
        Abort("DistributeVecOrigEq(): Internal error: TotCompnts != length") ;

#ifdef INFO
    /* Interchange roles of sends and receives for u  */
    PrintCom(P, l, COL, ComVol, MaxIn, MaxOut, MaxCompnts, TotCompnts) ;
#endif
  
    /*## Cleanup processor arrays: ##*/
    free(NprocsUV) ; /* no check needed */
    free(procindexV) ;
    free(procstartV) ;
    free(NprocsV) ;
    free(procindexU) ; 
    free(procstartU) ;
    free(NprocsU) ;
    free(SumsV) ;
    free(SumsU) ;
    free(NvV) ;
    free(NvU) ;
    free(NrV) ;
    free(NrU) ;
    free(NsV) ;
    free(NsU) ;
    free(ProcHistogramUV) ;
    free(ProcHistogramV) ;
    free(ProcHistogramU) ;
  
    return(maxcom) ;

} /* end DistributeVecOrigEq */
  
  
void InitIntersection(long l, int P,
                      long *procstartU, int *procindexU,
                      long *procstartV, int *procindexV, int *NprocsUV) {

    /* This function computes the intersection of the processor lists for
       row and column j, for each j, 0 <= j < l, and brings the processors
       of the intersection to the front in the list of row j.

       Input:  procindexU, contains the processor numbers occurring in
                       the matrix rows. The processor numbers for one
                       row j are stored consecutively.
               procstartU, where procstartU[j] is the position in array
                       procindexU of the first processor number of row j,
                       0 <= j < l.
                       procstartU[l] = total number of processor numbers
                       of all the rows together.

               procstartV, procindexV, the same for columns.

       Output: NprocsUV, where NprocsUV[j] is the number of processors that own 
                       nonzeros both in matrix row j and column j.

               procindexU is modified such that
                       procindexU[procstartU[j]..procstartU[j]+NprocsUV[j]-1]
                       contains the processors in the intersection for row/col j.
                       The remaining processors of row j are in
                       procindexU[procstartU[j]+NprocsUV[j]..procstartU[j+1]-1] 

               procindexV is not modified.  */

    int q, t, *Tmp ;
    long j, s, *Present ;

    Present = (long *) malloc(P * sizeof(long)) ;
    Tmp = (int *) malloc(P * sizeof(int)) ;
    if ( Present == NULL || Tmp == NULL )
        Abort("InitIntersection(): Not enough memory") ;

    /* Initialise boolean array */
    for(q=0 ; q<P ; q++)
        /* processor q is not present in current row */
        Present[q] = -1 ;

    for(j=0 ; j<l ; j++){
        /* Read processors of current row j */
        for(s=procstartU[j]; s<procstartU[j+1] ; s++){
            q = procindexU[s] ;
            /* mark processor q present in row j, using the time stamp j */
            Present[q] = j ; 
        }

        /* Read processors of column j and copy intersection into Tmp */
        t = 0;
        NprocsUV[j] = 0 ;
        for(s=procstartV[j]; s<procstartV[j+1] ; s++){ 
            q = procindexV[s] ; 
            if (Present[q] == j) {
                /* q is in intersection */
                Tmp[t] = q ;
                t++ ;
                NprocsUV[j]++ ;
                Present[q] = -1 ; /* reset time stamp */
            }
        }

        /* Copy remaining processors of row j into Tmp */
        for(s=procstartU[j]; s<procstartU[j+1] ; s++){ 
            q = procindexU[s] ; 
            if (Present[q] == j) {  
                /* q is in row j but not in the intersection */ 
                Tmp[t] = q ;
                t++ ;
            }
            /* no need to reset */
        }

        /* Copy processors from Tmp back into procindexU, in the new order */
        for(s=procstartU[j], t=0; s<procstartU[j+1] ; s++, t++)
            procindexU[s] = Tmp[t] ; 
    }

    free(Tmp) ;
    free(Present) ;

} /* end InitIntersection */


int FindProcLowestSum (long *procstart, int *procindex, long j, int npj,
                       long *SumsU, long *SumsV ) {

    /* This function returns the processor with the lowest total sum
       in row/column j, searching the first npj processors 
       procindex[procstart[j]..procstart[j]+npj-1]. */

    int proc, q ;
    long s, sum, min ;

    if (npj < 1)
        return INT_MIN ;

    proc = procindex[procstart[j]] ;
    if (npj == 1)
        return proc ;
    min = SumsU[proc] + SumsV[proc] ;

    for ( s = procstart[j]+1; s < procstart[j] + npj ; s++){
        q = procindex[s] ;
        sum = SumsU[q] + SumsV[q] ;
        if ( sum < min ) {
            proc = q ;
            min = sum ;
        }
    }
    
    return proc ; 

} /* end FindProcLowestSum */
