#include "Graph.h"

extern struct opts Options ;

struct biparthypergraph CreateNewBiPartHyperGraph(long NrVertices, long NrNets,
                                                  long NrPins,
                                                  int StoreMat, char MType) { 

    /* This function creates a new bipartitioned hypergraph
       with NrVertices vertices, NrNets nets (hyperedges),
       and NrPins pins (nonzeros, if the hypergraph represents a sparse matrix).

       If the boolean StoreMat is true, the numerical values of the pins
       (nonzeros) are stored as well:
           - a real part, if the matrix is not pattern-only, i.e. MType <> 'P';
           - an imaginary part if the matrix is complex, i.e. MType = 'C'.

       The function initialises all variables of the hypergraph to 0,
       sets all pointers to NULL, and it allocates memory to store the nets,
       vertices, and pins.
    */
  

    register long t ;
    struct biparthypergraph HG ;

    HG.NrVertices = NrVertices ;
    HG.NrNets = NrNets ;
    HG.NrPins = NrPins ;

    /* Initialise array pointers to NULL */
    HG.V = NULL ;
    HG.VtxMoveLog = NULL ;
    HG.OptPartVtx = NULL ;
    HG.Vtx2MatIndex = NULL ;
    HG.N = NULL ;
    HG.Net2MatIndex = NULL ;
    HG.VtxAdjncy = NULL ;
    HG.NetAdjncy = NULL ;
    HG.MatReValue = NULL ;
    HG.MatImValue = NULL ;

    /* Allocate memory for the vertices */
    if ( HG.NrVertices > 0 ) {
        HG.V = (struct vertex *) malloc(HG.NrVertices * sizeof(struct vertex)) ;
        HG.VtxMoveLog = (long *) malloc(HG.NrVertices * sizeof(long)) ;
        HG.OptPartVtx = (int *) malloc(HG.NrVertices * sizeof(int)) ;

        if ( HG.V == NULL || HG.VtxMoveLog == NULL || HG.OptPartVtx == NULL )
            Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;

        if ( StoreMat ) {
            HG.Vtx2MatIndex = (long *) malloc(HG.NrVertices * sizeof(long)) ;
    
            if ( HG.Vtx2MatIndex == NULL )
                Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;
        }
    }

    /* Allocate memory for the nets */
    if ( HG.NrNets > 0 ) {
        HG.N = (struct net *) malloc(HG.NrNets * sizeof(struct net)) ;

        if ( HG.N == NULL )
            Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;

        if ( StoreMat ) {
            HG.Net2MatIndex = (long *) malloc(HG.NrNets * sizeof(long)) ;
  
            if ( HG.Net2MatIndex == NULL )
                Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;
        }
    }

    /* Allocate memory for the pins */

    if ( HG.NrPins > 0 ) {
        HG.VtxAdjncy = (long *) malloc(HG.NrPins * sizeof(long)) ;
        HG.NetAdjncy = (long *) malloc(HG.NrPins * sizeof(long)) ;
        
        if ( HG.VtxAdjncy == NULL || HG.NetAdjncy == NULL )
            Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;

        if ( StoreMat ) {
            if (MType != 'P') {
                HG.MatReValue = (double *) malloc(HG.NrPins * sizeof(double)) ;
  
                if ( HG.MatReValue == NULL )
                    Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;
            }
            if (MType == 'C') {
                HG.MatImValue = (double *) malloc(HG.NrPins * sizeof(double)) ;
  
                if ( HG.MatImValue == NULL )
                    Abort("CreateNewBiPartHyperGraph(): Not enough memory") ;
            }
        }
    }

    /* Clear all values for the vertices */
    for ( t = 0 ; t < HG.NrVertices ; t++ ) {
        HG.V[t].vtxwgt = 0 ;
        HG.V[t].iStart = 0 ;
        HG.V[t].iEnd = 0 ;
        HG.V[t].partition = 0 ;
        HG.V[t].GBentry = NULL ;      
        HG.OptPartVtx[t] = 0 ;
        if ( StoreMat )
            HG.Vtx2MatIndex[t] = 0 ;
    }

    /* Clear all values for the nets */
    for ( t = 0 ; t < HG.NrNets ; t++ ) {
        HG.N[t].netwgt = 0 ;
        HG.N[t].iStartP0 = 0 ;
        HG.N[t].iStartP1 = 0 ;
        HG.N[t].iEnd = 0 ;
        HG.N[t].dir = ROW;
        if ( StoreMat )
            HG.Net2MatIndex[t] = 0 ;
    }

    /* Clear all values for the pins */
    for ( t = 0 ; t < HG.NrPins ; t++ ) {
        HG.VtxAdjncy[t] = 0 ;
        HG.NetAdjncy[t] = 0 ;
        if ( StoreMat ) {
            if (MType != 'P')
                HG.MatReValue[t] = 0.0 ;
            if (MType == 'C')
                HG.MatImValue[t] = 0.0 ;
        }
    }

    /* Initialize the gain buckets */
    HG.GBVtx[0].NrBuckets = HG.GBVtx[1].NrBuckets = 0 ;
    HG.GBVtx[0].Root = HG.GBVtx[1].Root = NULL ;

    /* Clear weights */
    HG.WeightP[0] = HG.WeightP[1] = 0 ;

    /* Clear comm, log, dir */
    HG.CurComm = HG.MinComm = HG.OptComm = 0 ;
    HG.CurVtxLog = HG.MinVtxLog = 0 ;
    HG.SplitDir = ROW ;

    return(HG) ;

} /* end CreateNewBiPartHyperGraph */


void DeleteBiPartHyperGraph(struct biparthypergraph *pHG) {

    /* This function deletes a bipartitioned hypergraph
       by freeing all its memory arrays and resetting all
       its variables to 0.

       The order of deallocation is the reverse of the order
       in which memory was allocated, to improve the chance
       of freeing large contiguous memory chunks.
    */

    ClearGainBucket(&(pHG->GBVtx[0])) ;
    ClearGainBucket(&(pHG->GBVtx[1])) ;
  
    if ( pHG->NrPins > 0 ) {
        if ( pHG->MatImValue != NULL )
            free(pHG->MatImValue) ;
        if ( pHG->MatReValue != NULL )
            free(pHG->MatReValue) ;
        if ( pHG->NetAdjncy != NULL )
            free(pHG->NetAdjncy) ;
        if ( pHG->VtxAdjncy != NULL )
            free(pHG->VtxAdjncy) ;
    }
  
    if ( pHG->NrNets > 0 ) {
        if ( pHG->Net2MatIndex != NULL )
            free(pHG->Net2MatIndex) ;
        if ( pHG->N != NULL )
            free(pHG->N) ;
    }
  
    if ( pHG->NrVertices > 0 ) {
        if ( pHG->Vtx2MatIndex != NULL )
            free(pHG->Vtx2MatIndex) ;
        if ( pHG->OptPartVtx != NULL )
            free(pHG->OptPartVtx) ;
        if ( pHG->VtxMoveLog != NULL )
            free(pHG->VtxMoveLog) ;
        if ( pHG->V != NULL )
            free(pHG->V) ;      
    }
  
    pHG->NrVertices = pHG->NrNets = pHG->NrPins = 0 ;
  
    /* Clear weights */
    pHG->WeightP[0] = pHG->WeightP[1] = 0 ;
  
    /* Clear comm, log, dir */
    pHG->CurComm = pHG->MinComm = pHG->OptComm = 0 ;
    pHG->CurVtxLog = pHG->MinVtxLog = 0 ;
    pHG->SplitDir = ROW ;

} /* end DeleteBiPartHyperGraph */


struct biparthypergraph SparseMatrix2BiPartHyperGraph(struct sparsematrix A, 
                                                      int dir) {

    /* This function translates a sparse matrix A into a hypergraph,
       which is returned as output.

       If the split direction dir = ROW, matrix rows become vertices
       of the hypergraph, and columns become nets.
       If the split direction dir = COL, matrix columns become vertices,
       and rows become nets.
       If the split direction dir = FINEGRAIN, matrix nonzeros become vertices,
       and rows as well as columns become nets.
 
       If the split strategy is local ratio, the direction given by
       the input parameter is ignored, and the direction is determined
       based on the aspect ratio between nonempty rows and nonempty columns.
       Only nonempty rows and columns are included in the hypergraph.

       The numerical values of the matrix  nonzeros are stored as well
       (with the pins of the hypergraph):
           - a real part, if the matrix is not pattern-only, i.e.
             A.MMTypeCode[2] <> 'P';
           - an imaginary part if the matrix is complex, i.e.
             A.MMTypeCode[2] = 'C'.

    */

    struct biparthypergraph HG ;  
    register long i, j, t ;

    long NrNeRows, NrNeCols, *nnzR, *nnzC, *Row2HGIndex, *Col2HGIndex,
         pin_index, vtxwgt, v=-1, n=-1, nC=-1 ;
    int *Diag ;

    /* Initialise relevant array pointers to NULL,
       to prevent compilation warnings. */
    HG.V = NULL ;
    HG.Vtx2MatIndex = NULL ;
    HG.N = NULL ;
    HG.Net2MatIndex = NULL ;
    HG.VtxAdjncy = NULL ;
    HG.NetAdjncy = NULL ;
    HG.MatReValue = NULL ;
    HG.MatImValue = NULL ;

    /* Allocate memory */
    nnzR = (long *) malloc(A.m * sizeof(long)) ;
    nnzC = (long *) malloc(A.n * sizeof(long)) ;
    Row2HGIndex = (long *) malloc(A.m * sizeof(long)) ;
    Col2HGIndex = (long *) malloc(A.n * sizeof(long)) ;
    Diag =  (int *) malloc(A.m * sizeof(int)) ;
    if ( nnzR == NULL || nnzC == NULL || Row2HGIndex == NULL ||
         Col2HGIndex == NULL || Diag == NULL )
        Abort("SparseMatrix2BiPartHyperGraph(): Not enough memory") ;

    /* Initialise */
    for ( i = 0 ; i < A.m ; i++ ) {
        nnzR[i] = 0 ;
        Row2HGIndex[i] = -1 ;
        Diag[i] = FALSE ;
    }
    for ( j = 0 ; j < A.n ; j++ ) {
        nnzC[j] = 0 ;
        Col2HGIndex[j] = -1 ;
    }

    /* Count the number of elements in each row/column of the matrix A. 
       Check whether each row contains a dummy or a regular diagonal element */
    
    NrNeRows = 0 ;
    NrNeCols = 0 ;
    for ( t = 0 ; t < A.NrNzElts ; t++ ) {
        i = A.i[t] ;
        j = A.j[t] ;
        if ( nnzR[i] == 0 )
            NrNeRows++ ; /* increment number of nonempty rows */
        nnzR[i]++ ;      /* increment number of nonzeros of row i */

        if ( nnzC[j] == 0 )
            NrNeCols++ ;
        nnzC[j]++ ;

        if ( i == j &&  A.m == A.n &&  
             (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' || A.MMTypeCode[3]=='H')) 
            Diag[i] = TRUE ; /* row i has a nonzero diagonal element */
    }

    /* If the SplitStrategy is localratio, set the direction,
       overriding the direction given on input  */
    if ( Options.SplitStrategy == LocalRatio ) {
        if ( NrNeRows > NrNeCols )   
            dir = ROW ;
        else if ( NrNeCols > NrNeRows )   
            dir = COL ;
        else if (Random1(0,1) == 0) /* random tie-breaking */
            dir = ROW ;
        else
            dir = COL;
#ifdef INFO2
        if ( dir == ROW )
            printf("**** Split in row direction \n") ;
        else
            printf("**** Split in col direction \n") ;
#endif
    }

    /* Create new bipartitioned hypergraph */
    if ( dir == ROW )
        HG = CreateNewBiPartHyperGraph(NrNeRows, NrNeCols, A.NrNzElts,
                                       TRUE, A.MMTypeCode[2]) ;
    else if ( dir == COL )
        HG = CreateNewBiPartHyperGraph(NrNeCols, NrNeRows, A.NrNzElts,
                                       TRUE, A.MMTypeCode[2]) ;
    else if ( dir == FINEGRAIN )
        HG = CreateNewBiPartHyperGraph(A.NrNzElts, NrNeRows + NrNeCols, 2*A.NrNzElts,
                                       TRUE, A.MMTypeCode[2]) ;

    /* Store the split direction */
    HG.SplitDir = dir ;

    if ( HG.SplitDir == ROW ) {
        /* Insert the nonempty rows of the matrix as vertices in the hypergraph and
           register the corresponding row number in Vtx2MatIndex */
        t = 0;
        pin_index = 0 ;
        for ( i = 0 ; i < A.m ; i++ ) {
            if ( nnzR[i] > 0 ) {
                HG.V[t].iStart = HG.V[t].iEnd = pin_index ;
                vtxwgt = nnzR[i] ;

                if ( A.dummy !=  NULL && A.dummy[i] )
                    vtxwgt-- ;

                if ( A.m == A.n && 
                    (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' ||
                     A.MMTypeCode[3]=='H') &&
                     Options.SymmetricMatrix_UseSingleEntry == SingleEntYes ) {
                    /* matrix is symmetric and each off-diagonal nonzero
                       has weight 2 */
                    if ( Diag[i] )
                        vtxwgt += (nnzR[i] -1) ;
                    else
                        vtxwgt += nnzR[i] ; 
                }

                HG.V[t].vtxwgt = vtxwgt ;
                pin_index += nnzR[i] ;
                HG.Vtx2MatIndex[t] = i ;
                Row2HGIndex[i] = t; /* temporarily register inverse mapping */ 
                t++ ;
            }
        }

        /* Insert the nonempty columns of the matrix as nets in the hypergraph and
           register the corresponding column number in Net2MatIndex */
        t = 0 ;
        pin_index = 0 ;
        for ( j = 0 ; j < A.n ; j++ ) {
            if ( nnzC[j] > 0 ) {
                HG.N[t].netwgt = 1 ; /* uniform netweights for the time being */
                HG.N[t].iStartP0 = HG.N[t].iStartP1 = HG.N[t].iEnd = pin_index ;
                HG.N[t].dir = COL;
                pin_index += nnzC[j] ;
                HG.Net2MatIndex[t] = j ;
                Col2HGIndex[j] = t;
                t++ ;
            }
        }
    } else if ( HG.SplitDir == COL ) {
        /* Insert the nonempty columns of the matrix as vertices in the hypergraph and
           register the corresponding column number in Vtx2MatIndex */
        t = 0;
        pin_index = 0 ;
        for ( j = 0 ; j < A.n ; j++ ) {
            if ( nnzC[j] > 0 ) {
                HG.V[t].iStart = HG.V[t].iEnd = pin_index ;
                
                if (A.MMTypeCode[0] == 'W' && A.NrColWeights > 0)
                    HG.V[t].vtxwgt = A.ColWeights[j] ;
                else {
                    vtxwgt = nnzC[j] ;
    
                    if ( A.dummy != NULL && A.dummy[j] )
                        vtxwgt-- ;
 
                    if ( A.m == A.n && 
                        (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' ||
                         A.MMTypeCode[3]=='H') &&
                         Options.SymmetricMatrix_UseSingleEntry == SingleEntYes){
                        if ( Diag[j] )
                            vtxwgt += (nnzC[j] -1) ;
                        else
                            vtxwgt += nnzC[j] ;
                    }
                    HG.V[t].vtxwgt = vtxwgt ;
                }  

                pin_index += nnzC[j] ;
                HG.Vtx2MatIndex[t] = j ;
                Col2HGIndex[j] = t; /* temporarily register inverse mapping */ 
                t++ ;
            }
        }

        /* Insert the nonempty rows of the matrix as nets in the hypergraph and
           register the corresponding row number in Net2MatIndex */
        t = 0;
        pin_index = 0 ;   
        for ( i = 0 ; i < A.m ; i++ ) {
            if ( nnzR[i] > 0 ) {
                HG.N[t].netwgt = 1 ; /* uniform netweights for the time being */
                HG.N[t].iStartP0 = HG.N[t].iStartP1 = HG.N[t].iEnd = pin_index ;
                HG.N[t].dir = ROW;
                pin_index += nnzR[i] ;
                HG.Net2MatIndex[t] = i ;
                Row2HGIndex[i] = t;
                t++ ;
            }
        }
    } else if (HG.SplitDir == FINEGRAIN ) {
        /* Insert the nonempty rows of the matrix as nets in the hypergraph and
           register the corresponding row number in Net2MatIndex */
        t = 0 ;
        pin_index = 0 ;
        for ( i = 0 ; i < A.m ; i++) {
            if ( nnzR[i] > 0 ) {
                HG.N[t].netwgt  = 1 ; /* uniform netweights for the time being */
                HG.N[t].iStartP0 = HG.N[t].iStartP1 = HG.N[t].iEnd = pin_index ;
                HG.N[t].dir = ROW;
                pin_index += nnzR[i] ;
                HG.Net2MatIndex[t] = i ;
                Row2HGIndex[i] = t;
                t++ ;
            }
        }
        /* Insert the nonempty columns of the matrix as nets in the hypergraph and
           register the corresponding column number in Net2MatIndex */
        for ( j = 0 ; j < A.n ; j++){
            if ( nnzC[j] > 0 ) {
                HG.N[t].netwgt  = 1 ; /* uniform netweights for the time being */
                HG.N[t].iStartP0 = HG.N[t].iStartP1 = HG.N[t].iEnd = pin_index ;
                HG.N[t].dir = COL;
                pin_index += nnzC[j] ;
                HG.Net2MatIndex[t] = j ;
                Col2HGIndex[j] = t;
                t++ ;
            }
        }
    
        /* Insert the nonzeros of the matrix as vertices in the hypergraph */
        pin_index = 0 ;
        for ( t = 0 ; t < A.NrNzElts ; t++ ) {
            HG.V[t].iStart = HG.V[t].iEnd = pin_index ;

            i = A.i[t] ;
            j = A.j[t] ;
            vtxwgt = 1; /* the default value */

            if ( i == j && A.dummy != NULL &&  A.dummy[i] )
                vtxwgt = 0 ;
            if ( i != j && A.m == A.n &&
                 (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' ||
                  A.MMTypeCode[3]=='H') &&
                 Options.SymmetricMatrix_UseSingleEntry == SingleEntYes)
                    /* matrix is symmetric and each off-diagonal nonzero has weight 2 */
                    vtxwgt = 2 ;

            HG.V[t].vtxwgt = vtxwgt ;
            pin_index += 2;
        }
    }



    /* Insert the nonzero elements of the matrix as pins in the 
       hypergraph and fill the adjacency lists */
    for ( t = 0 ; t < A.NrNzElts ; t++ ) {
        if ( HG.SplitDir == ROW ) {
            v =  Row2HGIndex[A.i[t]] ;
            n =  Col2HGIndex[A.j[t]] ;
        } else if ( HG.SplitDir == COL ) {
            v =  Col2HGIndex[A.j[t]] ;
            n =  Row2HGIndex[A.i[t]] ;
        } else if ( HG.SplitDir == FINEGRAIN ) {
            v = t; /* vertex = nonzero */
            n = Row2HGIndex[A.i[t]] ;  /* row-net */
            nC = Col2HGIndex[A.j[t]] ; /* column-net */
        }

        /* Store a numerical value with the vertex */
        if ( HG.SplitDir == ROW || HG.SplitDir == COL) {
            /* in the order of the adjacency lists */
            if (A.MMTypeCode[2] != 'P')
                HG.MatReValue[HG.V[v].iEnd] = A.ReValue[t] ; 
            if (A.MMTypeCode[2] == 'C')
                HG.MatImValue[HG.V[v].iEnd] = A.ImValue[t] ; 
         } else if ( HG.SplitDir == FINEGRAIN ) {
            /* in the order of the vertices */
            if (A.MMTypeCode[2] != 'P')
                HG.MatReValue[v] = A.ReValue[t] ; 
            if (A.MMTypeCode[2] == 'C')
                HG.MatImValue[v] = A.ImValue[t] ; 
         }
        
        /* Store the net with the vertex.
           This is the row-net in case of the finegrain direction */
        HG.VtxAdjncy[HG.V[v].iEnd++] = n ;
    
        /* Store the vertex with the net */
        HG.NetAdjncy[HG.N[n].iEnd++] = v ;

        if ( HG.SplitDir == FINEGRAIN ) {
            /* Store the column-net with the vertex */
            HG.VtxAdjncy[HG.V[v].iEnd++] = nC ;

            /* Store the vertex with the column-net */
            HG.NetAdjncy[HG.N[nC].iEnd++] = v ;
        }
    }

    /* Deallocate memory */
    free(Diag) ;
    free(Col2HGIndex) ;
    free(Row2HGIndex) ;
    free(nnzC) ;
    free(nnzR) ;

    return(HG) ;

} /* end SparseMatrix2BiPartHyperGraph */


void BiPartHyperGraph2SparseMatrix(struct biparthypergraph HG,
                                   struct sparsematrix T,
                                   long lo, long hi, long *mid) {

    /* This function translates a hypergraph HG into a subset of the nonzeros
       of the sparse matrix T. The subset is given by a range lo..hi.
       The nonzeros of this range are given new values for the row index i,
       column index j, and the associated numerical values.
       The remainder of T remains unchanged.

       The nonzeros of the range are split into two parts:
       lo..mid-1 belonging to part 0 (representing one processor)
       mid..hi belonging to part 1 (representing another processor).

       The nonzero arrays are filled starting from both sides simultaneously.

       If the split direction HG.SplitDir = ROW or COL, nonzeros are obtained
       from the net adjacency lists of the vertices.
       If the split direction HG.SplitDir = FINEGRAIN, each vertex becomes a 
       matrix nonzero.

       If present, the numerical values of the matrix nonzeros are retrieved
       from the hypergraph as well:
           - a real part, if the matrix is not pattern-only, i.e.
             T.MMTypeCode[2] <> 'P';
           - an imaginary part if the matrix is complex, i.e.
             T.MMTypeCode[2] = 'C'.
       This function works both when numerical values are stored and
       when they are not.

    */

    int P ;
    long new, n,
         left = lo,   /* lowest index of nonzero in range */ 
         right = hi ; /* highest index */
    register long t, tt ;
  
    for ( t = 0 ; t < HG.NrVertices ; t++ ) {
        P = HG.OptPartVtx[t] ; /* the part of vertex t in
                                  the optimal partitioning */
  
        if ( HG.SplitDir == ROW || HG.SplitDir == COL ) {
            /* Consider all nets adjacent to vertex t */
            for ( tt = HG.V[t].iStart ; tt < HG.V[t].iEnd ; tt++ ) {
                /* create a new nonzero for the current net tt */
                if ( P == 0 ) 
                    new = left ;
                else  /* P = 1 */
                    new = right ;
                if (new < lo || new > hi)
                    Abort("BiPartHyperGraph2SparseMatrix(): nonzero outside range") ;

                if ( HG.SplitDir == ROW ) { 
                    T.i[new] = HG.Vtx2MatIndex[t] ;
                    T.j[new] = HG.Net2MatIndex[HG.VtxAdjncy[tt]] ;
                } else if ( HG.SplitDir == COL ) {
                    T.j[new] = HG.Vtx2MatIndex[t] ;
                    T.i[new] = HG.Net2MatIndex[HG.VtxAdjncy[tt]] ;
                }

                if (HG.MatReValue != NULL )
                    T.ReValue[new] = HG.MatReValue[tt] ;
                if (HG.MatImValue != NULL )
                    T.ImValue[new] = HG.MatImValue[tt] ;

                if ( P == 0 )
                    left++ ;
                else
                    right-- ;
            }
        } else if (HG.SplitDir == FINEGRAIN) {
            if ( P == 0 )
                new = left ;
            else  /* P = 1 */
                new = right ;

            /* Create one new nonzero for the two nets adjacent to vertex t */
            if ( HG.V[t].iEnd - HG.V[t].iStart != 2 )
                Abort("BiPartHyperGraph2SparseMatrix(): corrupted finegrain hypergraph") ;

            for ( tt = HG.V[t].iStart ; tt < HG.V[t].iEnd ; tt++ ) {
                n = HG.VtxAdjncy[tt] ; 
                if ( HG.N[n].dir == ROW )
                    /* net is row-net */
                    T.i[new] = HG.Net2MatIndex[n] ;
                else /* column-net */
                    T.j[new] = HG.Net2MatIndex[n] ;
            }

            if (HG.MatReValue != NULL )
                T.ReValue[new] = HG.MatReValue[t] ;
            if (HG.MatImValue != NULL )
                T.ImValue[new] = HG.MatImValue[t] ;

            if ( P == 0 )
                left++ ;
            else
                right-- ;
        }
    }
    
    /* output splitting point */
    *mid = left ;
  
} /* end BiPartHyperGraph2SparseMatrix */
