#include "Mondriaan.h"

/* Global variable:  */
struct opts Options ;

int main(int argc, char **argv) {

    /* This function is the main function, called mondriaan. It partitions
       a sparse matrix, a dense input vector, and a dense output vector
       for the purpose of parallel sparse matrix-vector multiplication u = A*v.

       The function tries to minimise the total communication volume
       by suitably partitioning the matrix, while keeping the computation
       load imbalance withing a user-specified fraction epsilon.
       For the resulting communication volume, it tries to balance
       the communication work among the processors by suitably
       partitioning the vectors. The partitioning of the input and output
       vectors can either be done independently, or by imposing
       distr(u) = distr(v).

       The sparse matrix is read from file using the Matrix Market (MM) format.
       Its partitioning is written to file using a variant of the MM format,
       where all nonzeros belonging to the same processor are listed together.

       If the input file is called
           pi.mtx
       and 4 processors are required, the following output files are produced:
           pi.mtx-P4 containing the matrix distribution
           pi.mtx-v4                input vector distribution
           pi.mtx-u4                output vector distribution
           pi.mtx-C4                Cartesian submatrices corresponding
                                    to the matrix distribution
       Furthermore, useful information on the partitioning process
       and the result, such as the load imbalance and communication
       volume obtained, is written to standard output.

       The sparse matrix partitioner can also be used for partitioning
       hypergraphs. To do this, Mondriaan must be used in 1D column mode. 
       Matrix columns then represent hypergraph vertices; they can be given
       integer weights. Matrix rows then represent hyperedges; they cannot
       be given weights.

       The following additional operations are carried out by the main program:
           - timing the matrix and vector partitionings
           - removing duplicate nonzeros
           - (optionally) checking whether the matrix is structurally symmetric
           - (optionally) expanding a symmetric matrix stored in lower
             triangular format to a fully stored matrix
           - (optionally) adding diagonal dummy nonzeros to a square matrix
             to make the diagonal completely nonzero (thereby facilitating
             the vector distribution with distr(u) = distr(v)).

       All partitioning options and parameters can be set using
       the file mondriaan.defaults (provided the main function is compiled
       with the name mondriaan.) If no such file exists, it is created with
       sensible defaults.

       If the input matrix has already been partitioned, it will be
       partitioned again, from scratch.
 */

    struct sparsematrix A ; /* The Matrix ;-) */

    long MinNrNzElts, MaxNrNzElts ; /* minimum, maximum number of nonzero elements 
                                       per processor */
    double AvgNrNzElts ; /* average number of nonzero matrix elements per processor */

    long MinWeight, MaxWeight, TotWeight; /* minimum, maximum, total column weight */
    double AvgWeight ;                    /* average column weight */
    long weight, i ;

    long ComU, ComV ; /* communication cost for vectors u, v  */
    long q, nzq ;     /* processor number and its corresponding number of nonzeros */
    int weighted, symmetric ; /* boolean registering whether the input matrix
                                 was weighted, symmetric */

    int *u_proc, *v_proc ;   /* vector distribution for u, v */
    char output[MAX_WORD_LENGTH] ; /* filename of the output */ 
    
    /* Timing variables */
    clock_t starttime, endtime ;
    double cputime;
#ifdef UNIX
    struct timeval starttime1, endtime1 ;
#endif
  
    /* Get the parameters from the command line and initialise Options */
    GetParameters(&Options, argc, argv) ; 
  
    /* Initialise and read the matrix */
    A = MMReadSparseMatrix(Options.matrix) ;
  
#ifdef INFO
    printf("\n**************************************************************\n") ;
    printf("Problem statistics:\n") ;
    printf("  Matrix:           : %s\n",Options.matrix) ;
    printf("  %s %s %s %s %s\n", A.Banner,A.Object,A.Format,A.Field,A.Symmetry) ;
    printf("  m = Nr rows       : %ld\n", A.m) ; 
    printf("  n = Nr columns    : %ld\n", A.n) ; 
    printf("  nz = Nr nonzeros  : %ld\n", A.NrNzElts) ;
  
    /* Check if matrix A is structurally symmetric */
    if (SparseMatrixStructurallySymmetric(&A) )
        printf("  Matrix is structurally symmetric\n\n") ;
    else
        printf("  Matrix is structurally unsymmetric\n\n") ;
#endif

    /* Remove duplicate nonzeros by adding them */
    SparseMatrixRemoveDuplicates(&A) ;

    /* Check if matrix A is already distributed */
    if ( A.MMTypeCode[0] == 'D' ) {
        /* Matrix will be partitioned again */
        fprintf(stderr, "Warning: Matrix '%s' already distributed !\n", 
                Options.matrix) ;
        fprintf(stderr, "         (Ignoring current partitions)\n") ; 
        
        A.NrProcs = 0 ;
        if ( A.Pstart != NULL )
            free(A.Pstart) ;
        A.Pstart = NULL ;
    }
   
    /* Check if matrix is weighted (thus representing a hypergraph).
       In that case, it must have n column weights (representing vertex weights), 
       0 row weights and the split strategy must be onedimcol */

    if ( A.MMTypeCode[0] == 'W' && A.NrColWeights != A.n ) 
        Abort("Mondriaan(): weighted matrix with NrColWeights <> n\n") ;

    if ( A.MMTypeCode[0] == 'W' && A.NrRowWeights != 0) {
        fprintf(stderr, "Warning: Matrix '%s' has row weights!\n",
                Options.matrix) ;
        fprintf(stderr, "         Row-weighted column partitioning not yet implemented\n") ;
        fprintf(stderr, "         (Ignoring row weights)\n") ;
                
        A.NrRowWeights = 0 ;
        if ( A.RowWeights != NULL )
            free(A.RowWeights) ;
        A.RowWeights = NULL ;
    }

    if ( A.MMTypeCode[0] == 'W' && Options.SplitStrategy != OneDimCol) {
        fprintf(stderr, "Warning: Matrix '%s' is a weighted matrix!\n",
                Options.matrix) ;
        fprintf(stderr, "         must be partitioned by onedimcol strategy\n") ;
        fprintf(stderr, "         (Ignoring requested split strategy)\n") ;
        Options.SplitStrategy = OneDimCol ;
    }

    /* Register whether the input matrix was weighted, since the object type code will
       be changed by the partitioning procedure to the code 'D' for a distributed matrix */
    if ( A.MMTypeCode[0] == 'W' )
        weighted = TRUE ;
    else
        weighted = FALSE ;
  
    /* Register whether the input matrix was symmetric, since the symmetry type code will
       be changed by the conversion to full, to the code 'G' for a general matrix */
    if ( A.m == A.n && 
         (A.MMTypeCode[3]=='S' || A.MMTypeCode[3]=='K' || A.MMTypeCode[3]=='H')) {
        symmetric = TRUE ; 
    } else 
        symmetric = FALSE ;  
        
    if ( symmetric ) {
        if ( Options.SymmetricMatrix_UseSingleEntry == SingleEntNo )
            SparseMatrixSymmetric2Full(&A) ; 
        else if ( Options.SymmetricMatrix_SingleEntryType == ETypeRandom )
            SparseMatrixSymmetricLower2Random(&A) ;
    }
  
    /* If the matrix is square, add the dummies if requested.
       This may lead to an enhanced vector distribution in the case of
       an equal distribution of the input and output vectors.  */
    if ( A.m == A.n && 
         Options.SquareMatrix_DistributeVectorsEqual == EqVecYes &&
         Options.SquareMatrix_DistributeVectorsEqual_AddDummies == DumYes )
        AddDummiesToSparseMatrix(&A) ;
  
    /* Set the number of processors */
    A.NrProcs = Options.P ;
    
#ifdef INFO
    printf("  P = Nr processors : %d\n", A.NrProcs) ;
    printf("  Nr dummies added  : %ld\n", A.NrDummies) ;
    printf("  allowed imbalance : %g %c\n", 100*Options.eps, '%') ;
    printf("\n****** Mondriaan matrix distribution ******\n\n") ;
#endif

    /* Initialise Pstart with all nonzeros in processor 0 */
    A.Pstart = (long *) malloc((A.NrProcs+1) * sizeof(long)) ;
    if ( A.Pstart == NULL )
        Abort("Mondriaan(): Not enough memory for Pstart") ;
    A.Pstart[0] = 0 ;
    for ( q = 1 ; q <= A.NrProcs ; q++ )
        A.Pstart[q] = A.NrNzElts ;
  
    /**** Distribute the matrix (and time it) ****/
    starttime = clock() ;
#ifdef UNIX
    gettimeofday(&starttime1, NULL) ;
#endif
  
    DistributeMatrixMondriaan(&A, A.NrProcs, Options.eps) ; 
    
    endtime = clock() ;
#ifdef UNIX
    gettimeofday(&endtime1, NULL) ;
#endif
    cputime = ((double) (endtime - starttime)) / CLOCKS_PER_SEC ;
  
#ifdef TIME
    printf("  matrix distribution CPU-time    : %f seconds\n", cputime); 
#ifdef UNIX 
    printf("  matrix distribution elapsed time: %f seconds\n",
             (endtime1.tv_sec - starttime1.tv_sec) +
             (endtime1.tv_usec - starttime1.tv_usec) / 1000000.0 ) ; 
#endif 
    fflush(stdout) ;
#endif 
  
    /* Remove the dummies */
    if ( A.m == A.n &&
         Options.SquareMatrix_DistributeVectorsEqual == EqVecYes && 
         Options.SquareMatrix_DistributeVectorsEqual_AddDummies == DumYes )
        RemoveDummiesFromSparseMatrix(&A) ;
  
    /* Print information about the number of matrix elements */
    AvgNrNzElts = (double) A.NrNzElts / A.NrProcs ;
    MinNrNzElts = LONG_MAX ;
    MaxNrNzElts = LONG_MIN ;
    for ( q = 0 ; q < A.NrProcs ; q++ ) {
        nzq = A.Pstart[q+1] - A.Pstart[q] ;
        if ( nzq < MinNrNzElts )
            MinNrNzElts = nzq ;
        if ( nzq > MaxNrNzElts )
            MaxNrNzElts = nzq ;
    }

#ifdef INFO
    printf("  Nr nonzero matrix elements:\n") ;
    printf("    tot         : %ld \n", A.NrNzElts) ;
    printf("    avg = tot/P : %g  \n", AvgNrNzElts) ;
    printf("    max         : %ld \n", MaxNrNzElts) ;
    printf("    min         : %ld \n", MinNrNzElts) ;
    printf("    imbalance   : %g %c\n",
                100 * (MaxNrNzElts/AvgNrNzElts - 1.0), '%') ;
#endif

    if ( weighted ){
        A.MMTypeCode[0] = 'W' ; /* temporarily, needed for computing weights */
        TotWeight = ComputeWeight(A, 0, A.NrNzElts-1, NULL) ;
        AvgWeight = (double) TotWeight / A.NrProcs ;
        MinWeight = LONG_MAX ;
        MaxWeight = LONG_MIN ;
        for ( q = 0 ; q < A.NrProcs ; q++ ) {
            weight = ComputeWeight(A, A.Pstart[q], A.Pstart[q+1]-1, NULL) ;
            if ( weight < MinWeight )
                MinWeight = weight ;
            if ( weight > MaxWeight )
                MaxWeight = weight ;
        }
        A.MMTypeCode[0] = 'D' ; /* back to distributed matrix */

#ifdef INFO
        printf("  Vertex (column) weight:\n") ;
        printf("    tot         : %ld \n", TotWeight) ;
        printf("    avg = tot/P : %g  \n", AvgWeight) ;
        printf("    max         : %ld \n", MaxWeight) ;
        printf("    min         : %ld \n", MinWeight) ;
        printf("    imbalance   : %g %c\n",
                    100 * (MaxWeight/AvgWeight - 1.0), '%') ;
#endif
    }

    /* Convert randomly represented symmetric matrix to standard
       lower triangular form */
    if ( symmetric &&
         Options.SymmetricMatrix_UseSingleEntry == SingleEntYes &&
         Options.SymmetricMatrix_SingleEntryType == ETypeRandom )
              SparseMatrixSymmetricRandom2Lower(&A) ;
  
    /* Write the distributed matrix to file */
    sprintf(output, "%s-P%d", Options.matrix, A.NrProcs) ;  
    MMWriteSparseMatrix(A, output) ;
  
    /* Write the index sets of the Cartesian submatrices to file */
    sprintf(output, "%s-C%d", Options.matrix, A.NrProcs) ;
    MMWriteCartesianSubmatrices(A, output) ; 
  
    /* Convert symmetrically partitioned, symmetric matrix to full form 
       for vector distribution */
    if ( symmetric &&
         Options.SymmetricMatrix_UseSingleEntry == SingleEntYes ) {

        SparseMatrixSymmetric2Full(&A) ; /* now A.MMTypeCode[3]='G' */
        
        /* Print information about the number of matrix elements */
        AvgNrNzElts = (double) A.NrNzElts / A.NrProcs ;
        MinNrNzElts = LONG_MAX ;
        MaxNrNzElts = LONG_MIN ;
        for ( q = 0 ; q < A.NrProcs ; q++ ) {
             nzq = A.Pstart[q+1] - A.Pstart[q] ;
             if ( nzq < MinNrNzElts )
                 MinNrNzElts = nzq ;
             if ( nzq > MaxNrNzElts )
                 MaxNrNzElts = nzq ;
        }

#ifdef INFO
        printf("  After conversion of distributed symmetric matrix to"
               "  distributed full matrix\n" );
        printf("  Nr matrix elements:\n") ;
        printf("    tot         : %ld \n", A.NrNzElts) ;
        printf("    avg = tot/P : %g  \n", AvgNrNzElts) ;
        printf("    max         : %ld \n", MaxNrNzElts) ;
        printf("    min         : %ld \n", MinNrNzElts) ;

        printf("    imbalance   : %g %c\n", 
                    100 * (MaxNrNzElts / AvgNrNzElts - 1.0), '%') ;
        /* vertex weight information is not printed, since it is not
           meaningful here */ 
#endif   
    }
  
#ifdef INFO
    printf("\n****** Mondriaan vector distribution ******\n") ;
#endif
  
    /* Allocate memory for processor numbers for the vector components, 
       for u = A*v with A an m by n matrix */
    u_proc = (int *) malloc(A.m * sizeof(int)) ;
    v_proc = (int *) malloc(A.n * sizeof(int)) ;
    if ( u_proc == NULL || v_proc == NULL )
        Abort("Mondriaan(): Not enough memory for vectors u_proc and v_proc") ;
      
    /**** Distribute the vector (and time it) ****/
    starttime = clock() ;
#ifdef UNIX
    gettimeofday(&starttime1, NULL) ;
#endif
   
    if ( A.m == A.n && Options.SquareMatrix_DistributeVectorsEqual == EqVecYes ) {
        /* Distribute the vectors equally */
        if ( symmetric &&
             Options.SymmetricMatrix_UseSingleEntry == SingleEntYes ) {
             
            /* Distribute v independently, and then use the same distribution for u */
            ComV = DistributeVec(A, v_proc, ROW) ;
            for ( i = 0 ; i < A.m ; i++ )
                u_proc[i] = v_proc[i] ;
            ComU = ComV ;
#ifdef INFO
            printf("\nCommunication results for u are the same as for v\n") ;
            printf("but with sends and receives reversed\n\n") ;
#endif
        } else {
            /* Distribute u and v together */
            ComU = DistributeVecOrigEq(A, u_proc, v_proc) ;  
            ComV = 0 ;
        }
    } else {
        /* Distribute the vectors independently */
        ComV = DistributeVec(A, v_proc, ROW) ;
        ComU = DistributeVec(A, u_proc, COL) ;
    }
  
    endtime = clock() ;
#ifdef UNIX
    gettimeofday(&endtime1, NULL) ;
#endif
    cputime = ((double) (endtime - starttime)) / CLOCKS_PER_SEC ;
  
#ifdef TIME
    printf("  vector distribution CPU-time      : %f seconds\n", cputime);
#ifdef UNIX  
    printf("  vector distribution elapsed time  : %f seconds\n",
             (endtime1.tv_sec - starttime1.tv_sec) +
             (endtime1.tv_usec - starttime1.tv_usec) / 1000000.0 ) ; 
#endif   
    fflush(stdout) ;
#endif  
      
#ifdef INFO
    printf("\nCommunication cost for u = A v: %ld (%g) \n", ComU + ComV,
           ( (double) ComU / A.m  + (double) ComV / A.n ) /2 ) ;  
    fflush(stdout) ;
#endif
  
    /* Write the vector distribution to file */
    sprintf(output, "%s-u%d", Options.matrix, A.NrProcs) ;    
    WriteVectorDistribution(u_proc, A.m, A.NrProcs, output) ;
  
    sprintf(output, "%s-v%d", Options.matrix, A.NrProcs) ;    
    WriteVectorDistribution(v_proc, A.n, A.NrProcs, output) ;
  
    /* Free memory */
    MMDeleteSparseMatrix(&A) ;  
    free(v_proc) ;
    free(u_proc) ;

    /* Exit :) */
    exit(0) ;
    
} /* end main */
