#include "Options.h"

extern struct opts Options ;

/* This is the unique place in the Mondriaan package where the version number
   is defined. Date 9-7-2008 */
static char mversion[] = "2.0" ;

void GetParameters(struct opts *pOptions, int argc, char **argv) {

    /* This function gets the parameters from the command line
       and enters these into the Options data structure.
       The command line has the syntax
           program matrix P eps -option0=value0 -option1=value1 etc.
       For example:
           mondriaan pi.mtx 4 0.1 -SplitStrategy=onedimrow
       
       All parameters are first set to their default value,
       defined by file program.defaults, provided it exists.
       Otherwise this file is created.
       The optional parameters (option0, option1, etc) then overrule the 
       corresponding default values.

    */

    int t, i ;
    char c, *option, *value ;
    const char *eq = "=" ;
  
    /* Print Help if no option is given or a help option */
    if ( argc == 1 ||  ! strcmp(argv[1], "-h") ||
         ! strcmp(argv[1], "-help") || ! strcmp(argv[1], "--help") ) {
        PrintHelp(argc, argv) ;
        exit(1) ;
    }
  
    /* Check the number of arguments */
    if ( argc < 4 ) {
        printf("%s: invalid number of arguments \n", argv[0]) ;
        PrintHelp(argc, argv) ;
        exit(1) ;
    }    
  
    /* Read the required arguments */
    pOptions->argc = argc ;
    pOptions->argv = argv ;
    pOptions->matrix = argv[1] ;
    pOptions->P = atoi(argv[2]) ;
    if ( pOptions->P < 0 )
        Abort("GetParameters(): P out of range") ;
    pOptions->eps = atof(argv[3]) ;
    if ( pOptions->eps < 0 )
        Abort("GetParameters(): eps out of range") ;
  
    /* Set the default options */
    SetDefaults(pOptions) ;
    
    /* Read the optional command line arguments, e.g.
           -SplitStrategy=onedimrow
       to overrule the corresponding defaults */
    t = 4 ;
    while ( t < argc ) {
        c = argv[t][0] ;
        if ( c != '-')
            Abort("GetParameters(): option without -") ;

        /* Remove the minus sign from the option string */
        option = argv[t] + 1 ;
  
        if ( strchr(option, '=') == NULL ) { 
            /* No equal sign occurs in the option string.
               Use this argument as the option and the next argument
               as its value. Thus,
                   -SplitStrategy onedimrow
               will also work.  */
            t++ ;
            value = argv[t] ;
        } else { 
            /* Split option string into two strings: option, value */
            i = (int) strcspn(option, eq) ; /* index of the equal sign */

            /* Replace equal sign by end of string */
            option[i] = '\0' ;

            /* Set remainder becomes the value */
            value = option + i + 1 ;
        }  

        /* Set the current option to its proper value
           in the Options data structure */
        SetOption(pOptions, option, value) ;
        t++ ;
    }

} /* end GetParameters */
  
  
void SetDefaults(struct opts *pOptions) {

    /* This function sets the options to their default values
       using a defaults file, which can be hand-edited.
       If a defaults file (e.g. mondriaan.defaults) is already present
       it will be read and used. If no such file exists,
       it will be created and used immediately. 

       If a defaults file exists, but is corrupt (e.g. because an option
       is missing), then do not use it. Instead remove it,
       and let the program create a new file.

       Each line of the defaults file contains an option and its desired value.
       Empty lines and lines starting with a # are ignored. */

    FILE *fp ;
    int first, last ;

    char filename[MAX_WORD_LENGTH] ;
    char *line, linebuffer[MAX_LINE_LENGTH] ;
    char option[MAX_WORD_LENGTH], value[MAX_WORD_LENGTH] ;
  
    /* Set filename to program.defaults */
    strcpy(filename, pOptions->argv[0]) ;
    strcat(filename, ".defaults") ;
  
    /* If the defaults file does not exist, create it */
    if ( (fp = fopen(filename, "r")) == NULL ) {
        fprintf(stderr, "Cannot open file \"%s\".\n", filename) ;
        fprintf(stderr, "Creating file \"%s\" to store default options.\n", 
                filename) ;
  
        /* Write the options to file */
        fp = OpenFile(filename, "w") ;
        fprintf(fp, "# Mondriaan version %s default settings: \n", mversion) ;
        fprintf(fp, "LoadbalanceStrategy                            constant \n");
        fprintf(fp, "LoadbalanceAdjust                              yes \n") ;
        fprintf(fp, "SplitStrategy                                  localbest \n");
        fprintf(fp, "Alternate_FirstDirection                       ratio \n") ;
        fprintf(fp, "SplitMethod                                    KLFM \n");
        fprintf(fp, "SquareMatrix_DistributeVectorsEqual            no \n") ;
        fprintf(fp, "SquareMatrix_DistributeVectorsEqual_AddDummies yes \n") ;
        fprintf(fp, "SymmetricMatrix_UseSingleEntry                 no \n") ;
        fprintf(fp, "SymmetricMatrix_SingleEntryType                lower \n") ;
        fprintf(fp, "Coarsening_NrVertices                          200 \n") ;
        fprintf(fp, "Coarsening_MaxNrVtxInMatch                     2 \n") ;
        fprintf(fp, "Coarsening_StopRatio                           0.05 \n") ;
        fprintf(fp, "Coarsening_VtxMaxFractionOfWeight              0.2 \n") ;
        fprintf(fp, "Coarsening_MatchingStrategy                    inproduct \n");
        fprintf(fp, "Coarsening_InprodMatchingOrder                 decrwgt \n");
        fprintf(fp, "Coarsening_FineMatchingStrategy                ipfine \n");
        fprintf(fp, "Coarsening_NetScaling                          linear\n");
        fprintf(fp, "Coarsening_InprodScaling                       min\n");
        fprintf(fp, "Coarsening_MatchIdenticalFirst                 yes\n");
        fprintf(fp, "KLFM_InitPart_NrRestarts                       8 \n") ;
        fprintf(fp, "KLFM_InitPart_MaxNrLoops                       25 \n") ;
        fprintf(fp, "KLFM_InitPart_MaxNrNoGainMoves                 200 \n") ;
        fprintf(fp, "KLFM_Refine_MaxNrLoops                         25 \n") ;
        fprintf(fp, "KLFM_Refine_MaxNrNoGainMoves                   200 \n") ;
        fprintf(fp, "VectorPartition_Step3                          random \n") ;
        fprintf(fp, "VectorPartition_MaxNrLoops                     10 \n") ;
        fprintf(fp, "VectorPartition_MaxNrGreedyImproves            10 \n") ;
        fprintf(fp, "Seed                                           99 \n") ;
        CloseFile(filename, fp) ;
    }
  
    /* Read the option lines from the file */
    fp = OpenFile(filename, "r") ;
    while ( (line = fgets(linebuffer, MAX_LINE_LENGTH, fp)) != NULL ) { 
        /* a new line has been read */
        first = 0 ;
        last = strlen(line) ;
        if ( linebuffer[last] != '\0' )
            Abort("SetDefaults(): internal error") ;
  
        /* Remove initial blank characters from the line */
        while ( isspace((int) linebuffer[first]) ) 
            first++ ;
  
        line = linebuffer + first ;
  
        /* Clear option and value */
        option[0] = '\0' ;
        value[0]  = '\0' ;
  
        /* Read one option line. Ignore it if it is empty or starts with a # */
        if ( first != last && linebuffer[first] != '#' ) {
            /* Read option and value from the line */
            sscanf(line, "%s %s", option, value) ;
  
            /* Set the option */
            SetOption(pOptions, option, value) ;
        }
    }
    CloseFile(filename, fp) ;
  
    /* Check whether the numerical option values are within the proper range */
    if ( pOptions->Coarsening_NrVertices < 1 )
        Abort("SetDefaults(): Coarsening_NrVertices out of range") ;
  
    if ( pOptions->Coarsening_MaxNrVtxInMatch < 2 )
        Abort("SetDefaults(): Coarsening_MaxNrVtxInMatch out of range") ;    
  
    if ( pOptions->Coarsening_StopRatio < 0 || 
         pOptions->Coarsening_StopRatio > 1 )
        Abort("SetDefaults(): Coarsening_StopRatio out of range") ;
    
    if ( pOptions->Coarsening_VtxMaxFractionOfWeight <= 0 || 
         pOptions->Coarsening_VtxMaxFractionOfWeight > 1 )
        Abort("SetDefaults(): Coarsening_VtxMaxFractionOfWeight out of range") ;
  
    if ( pOptions->KLFM_InitPart_NrRestarts < 1 )
        Abort("SetDefaults(): KLFM_InitPart_NrRestarts out of range") ;
    
    if ( pOptions->KLFM_InitPart_MaxNrLoops < 1 )
        Abort("SetDefaults(): KLFM_InitPart_MaxNrLoops out of range") ;
    
    if ( pOptions->KLFM_InitPart_MaxNrNoGainMoves < 0 )
        Abort("SetDefaults(): KLFM_InitPart_MaxNrNoGainMoves out of range") ;
    
    if ( pOptions->KLFM_Refine_MaxNrLoops < 1 )
        Abort("SetDefaults(): KLFM_Refine_MaxNrLoops out of range") ;
  
    if ( pOptions->KLFM_Refine_MaxNrNoGainMoves < 0 )
        Abort("SetDefaults(): KLFM_Refine_MaxNrNoGainMoves out of range") ;
  
    if ( pOptions->VectorPartition_MaxNrGreedyImproves < 0 )
        Abort("SetDefaults(): VectorPartition_MaxNrGreedyImproves out of range") ;
  
    if ( pOptions->VectorPartition_MaxNrLoops < 1 )
        Abort("SetDefaults(): VectorPartition_MaxNrLoops out of range") ;
  
    /* Set the random number seed */
    if ( pOptions->Seed < -1 )
        Abort("SetDefaults(): Random number seed out of range") ;
    SetRandomSeed( pOptions->Seed ) ;

} /* end SetDefaults */
  

void SetOption(struct opts *pOptions, char *option, char *value) {

    /* This function sets the given option to its desired value
       in the Options data structure. The function does this
       by checking all the possible options to see whether
       the given option matches, and then checking all the
       possible values. */
        
  
    if ( ! strcmp(option, "LoadbalanceStrategy") ) {      
        if ( ! strcmp(value, "constant") )
            pOptions->LoadbalanceStrategy = Constant;
        else if ( ! strcmp(value, "increase") )
            pOptions->LoadbalanceStrategy = Increase ;
        else if ( ! strcmp(value, "decrease") )
            pOptions->LoadbalanceStrategy = Decrease ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "LoadbalanceAdjust") ) {
        if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
          pOptions->LoadbalanceAdjust = AdjustNo ;
        else if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
          pOptions->LoadbalanceAdjust = AdjustYes ;
        else { 
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "SplitStrategy") ) {      
        if ( ! strcmp(value, "alternate") )
            pOptions->SplitStrategy = Alternate ;
        else if ( ! strcmp(value, "localbest") )
            pOptions->SplitStrategy = LocalBest ;
        else if ( ! strcmp(value, "hybrid") )
            pOptions->SplitStrategy = Hybrid ;
        else if ( ! strcmp(value, "localratio") )
            pOptions->SplitStrategy = LocalRatio ;
        else if ( ! strcmp(value, "onedimrow") )
            pOptions->SplitStrategy = OneDimRow ;
        else if ( ! strcmp(value, "onedimcol") )
            pOptions->SplitStrategy = OneDimCol ;
        else if ( ! strcmp(value, "finegrain") )
            pOptions->SplitStrategy = FineGrain ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Alternate_FirstDirection") ) {      
        if ( ! strcmp(value, "row") )
            pOptions->Alternate_FirstDirection = Row ;
        else if ( ! strcmp(value, "col") )
            pOptions->Alternate_FirstDirection = Col ;
        else if ( ! strcmp(value, "ratio") )
            pOptions->Alternate_FirstDirection = Ratio ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "SplitMethod") ) {
        if ( ! strcmp(value, "simple") )
          pOptions->SplitMethod = Simple ;
        else if ( ! strcmp(value, "KLFM") || ! strcmp(value, "klfm") )
          pOptions->SplitMethod = KLFM ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "SquareMatrix_DistributeVectorsEqual") ) {
        if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
            pOptions->SquareMatrix_DistributeVectorsEqual = EqVecNo ;
        else if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
            pOptions->SquareMatrix_DistributeVectorsEqual = EqVecYes ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option,
                         "SquareMatrix_DistributeVectorsEqual_AddDummies") ) {
        if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
            pOptions->SquareMatrix_DistributeVectorsEqual_AddDummies = DumNo ;
        else if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
            pOptions->SquareMatrix_DistributeVectorsEqual_AddDummies = DumYes ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "SymmetricMatrix_UseSingleEntry") ) {
        if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
            pOptions->SymmetricMatrix_UseSingleEntry = SingleEntNo ;
        else if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
            pOptions->SymmetricMatrix_UseSingleEntry = SingleEntYes ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "SymmetricMatrix_SingleEntryType") ) {
        if ( ! strcmp(value, "lower") )
            pOptions->SymmetricMatrix_SingleEntryType = ETypeLower ;
        else if ( ! strcmp(value, "random") )
            pOptions->SymmetricMatrix_SingleEntryType = ETypeRandom ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Coarsening_NrVertices") ) {
        pOptions->Coarsening_NrVertices = atol(value) ;
    } else if ( ! strcmp(option, "Coarsening_MaxNrVtxInMatch") ) {
        pOptions->Coarsening_MaxNrVtxInMatch = atol(value) ;
    } else if ( ! strcmp(option, "Coarsening_StopRatio") ) {
        pOptions->Coarsening_StopRatio = atof(value) ;
    } else if ( ! strcmp(option, "Coarsening_VtxMaxFractionOfWeight") ) {
        pOptions->Coarsening_VtxMaxFractionOfWeight = atof(value) ;
    } else if ( ! strcmp(option, "Coarsening_MatchingStrategy") ) {
        if ( ! strcmp(value, "random") )
            pOptions->Coarsening_MatchingStrategy = Random ;
        else if ( ! strcmp(value, "inproduct") )
            pOptions->Coarsening_MatchingStrategy = Inprod ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Coarsening_FineMatchingStrategy") ) {
        if ( ! strcmp(value, "ip") )
            pOptions->Coarsening_FineMatchingStrategy = Ip ;
        else if ( ! strcmp(value, "ipfine") )
            pOptions->Coarsening_FineMatchingStrategy = IpFine ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Coarsening_InprodMatchingOrder") ) {
        if ( ! strcmp(value, "decrwgt") )
            pOptions->Coarsening_InprodMatchingOrder = DecreasingWgt ;
        else if ( ! strcmp(value, "incrwgt") )
            pOptions->Coarsening_InprodMatchingOrder = IncreasingWgt ;
        else if ( ! strcmp(value, "decrdeg") )
            pOptions->Coarsening_InprodMatchingOrder = DecreasingDegree ;
        else if ( ! strcmp(value, "incrdeg") )
            pOptions->Coarsening_InprodMatchingOrder = IncreasingDegree ;
        else if ( ! strcmp(value, "natural") )
            pOptions->Coarsening_InprodMatchingOrder = NaturalOrder ;
        else if ( ! strcmp(value, "random") )
            pOptions->Coarsening_InprodMatchingOrder = RandomOrder ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Coarsening_NetScaling") ) {
        if ( ! strcmp(value, "nonetscaling") || ! strcmp(value, "no") )
            pOptions->Coarsening_NetScaling = NoNetScaling;
        else if ( ! strcmp(value, "linear") )
            pOptions->Coarsening_NetScaling = Linear ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Coarsening_InprodScaling") ) {
        if ( ! strcmp(value, "noipscaling") || ! strcmp(value, "no") )
            pOptions->Coarsening_InprodScaling = NoIpScaling ;
        else if ( ! strcmp(value, "cos") )
            pOptions->Coarsening_InprodScaling = Cos ;
        else if ( ! strcmp(value, "min") )
            pOptions->Coarsening_InprodScaling = Min ;
        else if ( ! strcmp(value, "max") )
            pOptions->Coarsening_InprodScaling = Max ;
        else if ( ! strcmp(value, "jaccard") )
            pOptions->Coarsening_InprodScaling = Jaccard ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "Coarsening_MatchIdenticalFirst") ) {
        if ( ! strcmp(value, "no") || ! strcmp(value, "0") )
            pOptions->Coarsening_MatchIdenticalFirst = MatchIdNo ;
        else if ( ! strcmp(value, "yes") || ! strcmp(value, "1") )
            pOptions->Coarsening_MatchIdenticalFirst = MatchIdYes ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "KLFM_InitPart_NrRestarts") ) {
        pOptions->KLFM_InitPart_NrRestarts = atol(value) ;
    } else if ( ! strcmp(option, "KLFM_InitPart_MaxNrLoops") ) {
        pOptions->KLFM_InitPart_MaxNrLoops = atol(value) ;
    } else if ( ! strcmp(option, "KLFM_InitPart_MaxNrNoGainMoves") ) {
        pOptions->KLFM_InitPart_MaxNrNoGainMoves = atol(value) ;
    } else if ( ! strcmp(option, "KLFM_Refine_MaxNrLoops") ) {
        pOptions->KLFM_Refine_MaxNrLoops = atol(value) ;
    } else if ( ! strcmp(option, "KLFM_Refine_MaxNrNoGainMoves") ) {
        pOptions->KLFM_Refine_MaxNrNoGainMoves = atol(value) ;
    } else if ( ! strcmp(option, "VectorPartition_Step3") ) {
        if ( ! strcmp(value, "increase") )
            pOptions->VectorPartition_Step3 = VecIncrease ;
        else if ( ! strcmp(value, "decrease") )
            pOptions->VectorPartition_Step3 = VecDecrease ;
        else if ( ! strcmp(value, "random") )
            pOptions->VectorPartition_Step3 = VecRandom ;
        else {
            fprintf(stderr, "Unknown %s '%s'\n", option, value) ;
            Abort("SetOption(): incorrect default value") ;
        }
    } else if ( ! strcmp(option, "VectorPartition_MaxNrLoops") ) {
        pOptions->VectorPartition_MaxNrLoops = atol(value) ;
    } else if ( ! strcmp(option, "VectorPartition_MaxNrGreedyImproves") ) {
        pOptions->VectorPartition_MaxNrGreedyImproves  = atol(value) ;
    } else if ( ! strcmp(option, "Seed") ) {
        if ( ! strcmp(value, "random") )
            pOptions->Seed = -1 ;
        else
            pOptions->Seed = labs(atol(value)) ; 
    } else {
        fprintf(stderr, "Unknown option '%s' \n", option) ;
        Abort("SetOption(): incorrect option") ;
    }
    return ;

} /* end SetOption */
  
  
void PrintHelp(int argc, char **argv) {

    /* This function prints a help message on how to use Mondriaan
       from the command line. This function can for instance be called
       in case there are too few arguments. */

    printf("\nMondriaan version %s.\n", mversion) ;
    printf("Usage: %s [matrix] [P] [eps] <options>\n\n", argv[0]) ;
    printf(" [matrix] - the matrix to partition\n") ;
    printf(" [P]      - the number of processors\n") ;
    printf(" [eps]    - the maximum allowed load imbalance\n\n") ;
    printf(" <options>:") ;
    printf(" for details, see the User's Guide at\n") ;
    printf("            http://www.math.uu.nl/people/bisseling") ;
    printf("/Mondriaan/users_guide.html\n") ;
    printf("\n") ;
    fflush(stdout) ;

} /* end PrintHelp */
