#include "SparseMatrix.h"

void MMSparseMatrixInit(struct sparsematrix *pM) {

    /* This function initialises a sparse matrix by setting all its
       numerical variables to 0, all character variables to a blank,
       all string variables to the empty string,
       and all pointers to NULL. It does not allocate memory. */

    /* Initialise typeparams */
    pM->MMTypeCode[0] = ' ' ;
    pM->MMTypeCode[1] = ' ' ;
    pM->MMTypeCode[2] = ' ' ;
    pM->MMTypeCode[3] = ' ' ;

    /* Initialise strings */
    pM->Banner[0] = '\0' ;
    pM->Object[0] = '\0' ;
    pM->Format[0] = '\0' ;
    pM->Field[0] = '\0' ;
    pM->Symmetry[0] = '\0' ;

    /* Initialise matrix dimensions */
    pM->NrNzElts = 0 ;  /* no nonzeros yet */
    pM->m = 0 ;
    pM->n = 0 ;
    pM->NrDummies = 0 ; /* no dummies added yet */
    pM->NrProcs = 0 ;   /* matrix has not been partitioned yet */

    /* Initialise matrix arrays of length NrNzElts */
    pM->i = NULL ;
    pM->j = NULL ;
    pM->ReValue = NULL ;
    pM->ImValue = NULL ;

    /* Initialise other matrix arrays */
    pM->dummy = NULL ;
    pM->Pstart = NULL ;
    pM->header = NULL ;
    pM->tail = NULL ;

    /* Initialise weight information */
    pM->NrRowWeights = 0 ;
    pM->NrColWeights = 0 ;
    pM->RowWeights = NULL ;
    pM->ColWeights = NULL ;

    /* Initialise marker arrays */
    pM->RowMark = NULL ;
    pM->ColMark = NULL ;

} /* end MMSparseMatrixInit */
  

struct sparsematrix MMReadSparseMatrix(char *filename) {

    /* This function reads the header and the nonzeros of a sparse
       matrix in Matrix Market format, and possibly a tail.
       The header is an optional banner line
           %%MatrixMarket ...
       followed by a line
           m n nz
       giving the number of rows, columns, nonzeros.

       If the matrix has been partitioned (distributed-matrix),
       the function reads distribution information.
       In that case the banner line is followed by a line
           m n nz P
       where P is the number of processors, and then by
       P+1 lines giving the start of the nonzeros of the processors,
       where the (P+1)-th line equals nz.

       If the matrix is weighted (weighted-matrix),
       the function reads m integer weights for the rows
       and then n weights for the columns. */

    struct sparsematrix M ;
    FILE *fp ;
  
    MMSparseMatrixInit(&M) ;
    
    /* Open the file */
    fp = OpenFile(filename, "r") ;
  
    /* Read the header */
    MMSparseMatrixReadHeader(&M, fp) ;
  
    /* Allocate memory for the arrays of the matrix */
    MMSparseMatrixAllocateMemory(&M) ;
  
    /* Read the distribution information */
    MMSparseMatrixReadPstart(&M, fp) ;
  
    /* Read the nonzero entries of the matrix */
    MMSparseMatrixReadEntries(&M, fp) ;
  
    /* Read the row and column weights */
    MMSparseMatrixReadWeights(&M, fp) ;
  
    /* Read the tail */
    MMSparseMatrixReadTail(&M, fp) ;
  
    /* Close the file */
    CloseFile(filename, fp) ;  
  
    return(M) ;

} /* end MMReadSparseMatrix */
  


void MMWriteSparseMatrix(struct sparsematrix M, char *filename) {

    /* This function prints the header and the nonzeros of a sparse
       matrix in Matrix Market format, and possibly a tail,
       into a file.
       
       The header is an optional banner line
           %%MatrixMarket ...
       followed by a line
           m n nz
       giving the number of rows, columns, nonzeros.

       If the matrix has been partitioned (distributed-matrix),
       the function prints distribution information.
       In that case the banner line is followed by a line
           m n nz P
       where P is the number of processors, and then by
       P+1 lines giving the start of the nonzeros of the processors,
       where the (P+1)-th line equals nz.

       If the matrix is weighted (weighted-matrix),
       the function prints m integer weights for the rows
       and then n weights for the columns. */ 

    FILE *fp ;
  
    /* Open the file */
    fp = OpenFile(filename, "w") ;
  
    /* Print the header */
    MMSparseMatrixPrintHeader(&M, fp) ;
  
    /* Print the distribution information */
    MMSparseMatrixPrintPstart(&M, fp) ;
  
    /* Print the nonzero elements */
    MMSparseMatrixPrintEntries(&M, fp) ;
  
    /* Print the row and column weights */
    MMSparseMatrixPrintWeights(&M, fp) ;  
  
    /* Print the tail */
    MMSparseMatrixPrintTail(&M, fp) ;

    /* Close the file */
    CloseFile(filename, fp) ;

  } /* end MMWriteSparseMatrix */

  
void MMDeleteSparseMatrix(struct sparsematrix *pM) {

    /* This function frees the memory for all the arrays of the matrix M,
       and reinitialises the variables so the matrix can be reused. 
       It sets the pointers to all arrays to NULL. */
  
    MMSparseMatrixFreeMemory(pM) ;
    MMSparseMatrixInit(pM) ;

} /* end MMDeleteSparseMatrix */
  
  
void MMSparseMatrixSetTypeCode(struct sparsematrix *pM, char *line) {

    /* This function reads a Matrix Market banner line,
       splits it into 5 strings (for banner, object, format, field, symmetry),
       reads these into the corresponding strings of the matrix M,
       and sets the type codes accordingly. 

       The first string is the banner string. The possibilities for the
       other strings are:

       Object = matrix, distributed matrix, or weighted matrix 
       TypeCode[0] = M, D, W

       Format = coordinate, array
       TypeCode[1] = C, A

       Field =  integer, real, complex, or pattern
       TypeCode[2] = I, R, C, P

       Symmetry = general, symmetric, skew-symmetric, or hermitian
       TypeCode[3] = G, S, K, H

       If there are less than 5 strings in the line,
       or the first string differs from the Matrix Market banner,
       the four type codes are set to the defaults M, C, R, G, respectively,
       and the matrix strings are initialised accordingly.  */

    char *c ;
  
    /* Read Matrix Market Type Codes from line */
    if (sscanf(line, "%s %s %s %s %s", pM->Banner, pM->Object, 
        pM->Format, pM->Field, pM->Symmetry) != 5 ||
        strcmp(pM->Banner, MM_Banner) ) { 

        /* No correct Type Code found, assume default */
        strcpy(pM->Banner, MM_Banner) ;
        pM->MMTypeCode[0] = 'M' ; strcpy(pM->Object, "matrix") ;
        pM->MMTypeCode[1] = 'C' ; strcpy(pM->Format, "coordinate") ;
        pM->MMTypeCode[2] = 'R' ; strcpy(pM->Field, "real") ;
        pM->MMTypeCode[3] = 'G' ; strcpy(pM->Symmetry, "general") ;
        return ;
    }
  
    /*** Set Object type in M.MMTypeCode[0] ***/

    /* convert string to lower case */
    for ( c = pM->Object ; *c != '\0' ; c++)
        *c = tolower(*c) ;
  
    /* set Object type = matrix, distributed matrix, or weighted matrix */
    if ( ! strcmp(pM->Object, "matrix") )
        /* pM->Object = "matrix" */
        pM->MMTypeCode[0] = 'M' ;
    else if ( ! strcmp(pM->Object, "distributed-matrix") ||
              ! strcmp(pM->Object, "distributedmatrix") )
        pM->MMTypeCode[0] = 'D' ;      
    else if ( ! strcmp(pM->Object, "weighted-matrix") ||
              ! strcmp(pM->Object, "weightedmatrix") )
        pM->MMTypeCode[0] = 'W' ;      
    else {
        fprintf(stderr, "Unknown Object '%s'\n", pM->Object) ;
        Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }
    
    /*** Set Format type in M.MMTypeCode[1] ***/

    for ( c = pM->Format ; *c != '\0' ; c++)
        *c = tolower(*c) ;
  
    /* set Format type = coordinate or array */
    if ( ! strcmp(pM->Format, "coordinate") || 
         ! strcmp(pM->Format, "sparse") )
        pM->MMTypeCode[1] = 'C' ;
    else if ( ! strcmp(pM->Format, "array") || 
              ! strcmp(pM->Format, "dense") )
        pM->MMTypeCode[1] = 'A' ;
    else {
        fprintf(stderr, "Unknown Format '%s'\n", pM->Format) ;
        Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }
    
    /*** Set Field type in M.MMTypeCode[2] ***/

    for ( c = pM->Field ; *c != '\0' ; c++)
        *c = tolower(*c) ;
  
    /* set Field type = integer, real, complex, or pattern */
    if ( ! strcmp(pM->Field, "integer") )
        pM->MMTypeCode[2] = 'I' ;
    else if ( ! strcmp(pM->Field, "real") || 
              ! strcmp(pM->Field, "float") )
        pM->MMTypeCode[2] = 'R' ;
    else if ( ! strcmp(pM->Field, "complex") )
        pM->MMTypeCode[2] = 'C' ;
    else if ( ! strcmp(pM->Field, "pattern") )
        pM->MMTypeCode[2] = 'P' ;
    else {
        fprintf(stderr, "Unknown Field '%s'\n", pM->Field) ;
        Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }
  
    /*** Set Symmetry type in M.MMTypeCode[3] ***/

    for ( c = pM->Symmetry ; *c != '\0' ; c++)
        *c = tolower(*c) ;
  
    /* set Symmetry type = general, symmetric, skew-symmetric, or hermitian */
    if ( ! strcmp(pM->Symmetry, "general") )
        pM->MMTypeCode[3] = 'G' ;
    else if ( ! strcmp(pM->Symmetry, "symmetric") )
        pM->MMTypeCode[3] = 'S' ;
    else if ( ! strcmp(pM->Symmetry, "skew-symmetric") ||
              ! strcmp(pM->Symmetry, "skewsymmetric") )
        pM->MMTypeCode[3] = 'K' ;
    else if ( ! strcmp(pM->Symmetry, "hermitian") )
        pM->MMTypeCode[3] = 'H' ;
    else {
        fprintf(stderr, "Unknown Symmetry '%s'\n", pM->Symmetry) ;
        Abort("MMSparseMatrixSetTypeCode(): incorrect type") ;
    }

} /* end MMSparseMatrixSetTypeCode */
  

void MMSparseMatrixGetTypeCode(struct sparsematrix *pM, char *line) {

    /* This function reads the type codes from a matrix M,
       converts these into a Matrix Market banner line
       and copies this line to line.

       Furthermore, the 5 strings of the generated banner line
       (for banner, object, format, field, symmetry)
       are copied into the corresponding strings of the matrix M. 
 
       The type codes are converted into strings as follows:

       TypeCode[0] = M, D, W
       Object = matrix, distributed matrix, or weighted matrix

       TypeCode[1] = C, A
       Format = coordinate, array

       TypeCode[2] = I, R, C, P
       Field =  integer, real, complex, or pattern

       TypeCode[3] = G, S, K, H
       Symmetry = general, symmetric, skew-symmetric, or hermitian

    */

    /* Set Banner */
    strcpy(pM->Banner, MM_Banner) ;
  
    /* Set Object type */
    if ( pM->MMTypeCode[0] == 'M' )
        strcpy(pM->Object, "matrix") ;
    else if ( pM->MMTypeCode[0] == 'D' )
        strcpy(pM->Object, "distributed-matrix") ;
    else if ( pM->MMTypeCode[0] == 'W' )
        strcpy(pM->Object, "weighted-matrix") ;
    else {
        fprintf(stderr, "Unknown Object Type Code '%c'\n", pM->MMTypeCode[0]) ;
        Abort("MMSparseMatrixGetTypeCode(): incorrect type param") ;
    }
  
    /* Set Format type */
    if ( pM->MMTypeCode[1] == 'C' )
        strcpy(pM->Format, "coordinate") ;
    else if ( pM->MMTypeCode[1] == 'A' )
        strcpy(pM->Format, "array") ;
    else {
        fprintf(stderr, "Unknown Format Type Code '%c'\n", pM->MMTypeCode[1]) ;
        Abort("MMSparseMatrixGetTypeCode(): incorrect type param") ;
    }
  
    /* Set Field type */
    if ( pM->MMTypeCode[2] == 'I' )
        strcpy (pM->Field, "integer") ;
    else if ( pM->MMTypeCode[2] == 'R' )
        strcpy(pM->Field, "real") ;
    else if ( pM->MMTypeCode[2] == 'C' )
        strcpy(pM->Field, "complex") ;
    else if ( pM->MMTypeCode[2] == 'P' )
        strcpy(pM->Field, "pattern") ;
    else {
        fprintf(stderr, "Unknown Field Type Code '%c'\n", pM->MMTypeCode[2]) ;
        Abort("MMSparseMatrixGetTypeCode(): incorrect type param") ;
    }
  
    /* Set Symmetry type */
    if ( pM->MMTypeCode[3] == 'G' )
        strcpy(pM->Symmetry, "general") ;
    else if ( pM->MMTypeCode[3] == 'S' )
        strcpy(pM->Symmetry, "symmetric") ;    
    else if ( pM->MMTypeCode[3] == 'K' )
        strcpy(pM->Symmetry, "skew-symmetric")  ;
    else if ( pM->MMTypeCode[3] == 'H' )
        strcpy(pM->Symmetry, "hermitian") ;
    else {
        fprintf(stderr, "Unknown Symmetry Type Code '%c'\n", pM->MMTypeCode[3]) ;
        Abort("MMSparseMatrixGetTypeCode(): incorrect type param") ;
    }
  
    /* Copy the type parameters to the banner line */
    sprintf(line, "%s %s %s %s %s\n", pM->Banner, pM->Object, pM->Format, 
            pM->Field, pM->Symmetry) ;

} /* end MMSparseMatrixGetTypeCode */
  
  
void MMSparseMatrixAllocateMemory(struct sparsematrix *pM) {

    /* This function allocates memory for the following arrays of the matrix M:
           - the nonzero arrays (i, j, Val) 
           - Pstart, which represents the start of the nonzeros
             of the processors, for a distributed matrix
           - the row and column weights, for a weighted matrix.

       It does not allocate arrays for dummies or row/column marks,
       since these are only allocated where needed, by other functions. */

    /* Allocate memory for indices. Allocate one element extra,
       in case the matrix is empty. */
    pM->i = (long *) malloc((pM->NrNzElts+1) * sizeof(long)) ;
    pM->j = (long *) malloc((pM->NrNzElts+1) * sizeof(long)) ;
    if ( pM->i == NULL || pM->j == NULL )
        Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
              "for matrix indices") ;
  
    /* Allocate memory for numerical values */
    if ( pM->MMTypeCode[2] != 'P' ) { /* non-pattern matrix */
        pM->ReValue = (double *) malloc((pM->NrNzElts+1) * sizeof(double)) ;
        if ( pM->ReValue == NULL )
            Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
                  "for matrix values") ;
    }
    if ( pM->MMTypeCode[2] == 'C' ) { /* complex matrix */
        pM->ImValue = (double *) malloc((pM->NrNzElts+1) * sizeof(double)) ;
        if ( pM->ImValue == NULL )
            Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
                  "for matrix values") ;
    }
  
    /* Allocate memory for Pstart, for distributed matrix */
    if ( pM->MMTypeCode[0] == 'D' ) {
        pM->Pstart = (long *) malloc((pM->NrProcs+1) * sizeof(long)) ;
        if ( pM->Pstart == NULL )
            Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
                  "for Pstart") ;
    }
  
    /* Allocate memory for row weights, for weighted matrix */
    if ( pM->MMTypeCode[0] == 'W' && pM->NrRowWeights > 0 ) {
        pM->RowWeights = (long *) malloc(pM->NrRowWeights * sizeof(long)) ;
        if ( pM->RowWeights == NULL )
            Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
                  "for RowWeights") ;
    }
  
    /* Allocate memory for column weights, for weighted matrix */
    if ( pM->MMTypeCode[0] == 'W' && pM->NrColWeights > 0 ) {
        pM->ColWeights = (long *) malloc(pM->NrColWeights * sizeof(long)) ;
        if ( pM->ColWeights == NULL )
            Abort("MMSparseMatrixAllocateMemory(): Not enough memory "
                  "for ColWeights") ;
    }

} /* end MMSparseMatrixAllocateMemory */
  

void MMSparseMatrixFreeMemory(struct sparsematrix *pM) {

    /* This function frees memory for all the arrays of the matrix M:
           - the nonzero arrays (i, j, Val)
           - Pstart, which represents the start of the nonzeros
             of the processors, for a distributed matrix
           - the row and column weights, for a weighted matrix
           - dummies
           - row/column marks. */

    if ( pM->i != NULL )
        free(pM->i) ;
  
    if ( pM->j != NULL )
        free(pM->j) ;
  
    if ( pM->ReValue != NULL )
        free(pM->ReValue) ;
    
    if ( pM->ImValue != NULL )
        free(pM->ImValue) ;
  
    if ( pM->dummy != NULL )
        free(pM->dummy) ;
  
    if ( pM->Pstart != NULL )
        free(pM->Pstart) ;
  
    if ( pM->header != NULL )
        free(pM->header) ;
  
    if ( pM->tail != NULL )
        free(pM->tail) ;

    if ( pM->RowWeights != NULL )
        free(pM->RowWeights) ;
  
    if ( pM->ColWeights != NULL )
        free(pM->ColWeights) ;

    if ( pM->RowMark != NULL )
        free(pM->RowMark) ;
  
    if ( pM->ColMark != NULL )
        free(pM->ColMark) ;
  
} /* end MMSparseMatrixFreeMemory */
  
  
void MMSparseMatrixReadHeader(struct sparsematrix *pM, FILE *fp) {

    /* This function reads the header from the input stream fp.
       The header is an optional set of lines starting with %,
       (the first is a banner line, and the remainder are comment lines),
       followed by a line with the size of the matrix,
           m n nz
       where m = the number of rows, m >=1
             n = the number of columns, n >= 1
             nz the number of nonzeros, nz >= 1.
       The size is read into variables M.m, M.n, M.NrNzElts.
       
       The banner line sets the type of the matrix.
       If no such line is present the type is set to the default. 
       The comment lines are read into the string M.header.

       For a distributed matrix, the size line has an extra parameter:
           m n nz P
       where P is the number of processors, P >= 1.

       For a weighted matrix, the size line is
           m n nz param
       where param = 0 means no row or column weights
                     1       only row weights
                     2       only column weights
                     3       both row and column weights.
       
       This function removes initial white space characters
       (blanks, tabs) from each line, and empty lines,
       before reading the actual input. */

    int first, last, SizeRead=FALSE, TypeSet=FALSE, count=0, param=0 ;
    char *line, linebuffer[MAX_LINE_LENGTH] ; 
    long hsize ;
  
    pM->header = NULL ;
    hsize = 1 ;  /* header comment size (in characters) */
  
    while ( SizeRead == FALSE && 
           (line = fgets(linebuffer, MAX_LINE_LENGTH, fp)) != NULL ) { 
        /* a new line has been read */
        first = 0 ;
        last = strlen(line) ;
        if ( linebuffer[last] != '\0' )
            Abort("MMSparseMatrixReadTail(): internal error") ;

        /* Remove initial blank characters from the line */
        while ( isspace((int) linebuffer[first]) ) 
            first++ ;
  
        line = linebuffer + first ;
  
        if ( first != last ) {
            if ( linebuffer[first] == '%' ) {
                /* The first part of the header, a line starting with % */
                if ( TypeSet == FALSE ) {
                    /* Line is a banner line. Set the type using the banner,
                       or set it to the default */
                    MMSparseMatrixSetTypeCode(pM, line) ;
                    TypeSet= TRUE ;
                } else {
                    /* Add this line to the header comment */
                    hsize += last - first ;
                    pM->header = (char *) realloc(pM->header, hsize * 
                                                  sizeof(char)) ;
                    if ( pM->header == NULL )
                        Abort("ReadHeader(): Not enough memory for realloc "
                              "of matrix header") ;
                    pM->header[hsize + first - last - 1] = '\0' ;
                    strncat(pM->header, line, last-first) ;
                }
            } else {
                /* The second part of the header, the size line */

                /* Read m, n, nz, and perhaps an extra parameter
                   from the line */
                count = sscanf(line, "%ld %ld %ld %d\n",
                               &(pM->m), &(pM->n), &(pM->NrNzElts), &param) ;
        
                if ( count < 3 || pM->m < 1 || pM->n < 1 || pM->NrNzElts < 1 )
                    Abort("ReadHeader(): Error in matrix size") ;
                else
                    SizeRead = TRUE ;
            }
        }
    }
  
    /* If no type code has been set, set it to the default.
       This is needed if there are no lines starting with % */
    if ( TypeSet == FALSE ) 
        MMSparseMatrixSetTypeCode(pM, "") ;
  
    if ( pM->MMTypeCode[0] == 'D' ){
        /* distributed matrix: param = number of processors */ 
        if (count == 4 && param >= 1 )
            pM->NrProcs = param;
        else 
            Abort("ReadHeader(): Error in number of processors");

    } else if ( pM->MMTypeCode[0] == 'W' ){
        /* weighted matrix: param determines presence of row/column weights */

        if ( count < 4 || param < 0 || param >= 4 )
            Abort("ReadHeader(): Error in weight param");

        if ( param == 1 || param == 3 )
            pM->NrRowWeights= pM->m;
        else
            pM->NrRowWeights= 0; 

        if ( param == 2 || param == 3 ) 
            pM->NrColWeights= pM->n;
        else
            pM->NrColWeights= 0;
    }

} /* end MMSparseMatrixReadHeader */

  
void MMSparseMatrixReadPstart(struct sparsematrix *pM, FILE *fp) {

    /* This function reads the start of the nonzeros for each of the
       P processor parts, where P = M.NrProcs. The nonzeros of
       processor q start at position Pstart[q] and end at Pstart[q+1]-1,
       for 0 <= q < P.
       For convenience, Pstart[0] = 0 and Pstart[P] = nz(M). 
       This is checked, and if it does not hold the function aborts.
       Each Pstart[q] value should be on a separate line. */
       
    int t ;
  
    if ( pM->MMTypeCode[0] == 'D' ){
        for ( t = 0 ; t <= pM->NrProcs ; t++  )
            if ( fscanf(fp, "%ld\n", &(pM->Pstart[t])) != 1 )
                Abort("MMSparseMatrixReadPstart(): Pstart: Read Error") ;

        if ( pM->Pstart[0] != 0 || pM->Pstart[pM->NrProcs] != pM->NrNzElts )
            Abort("MMSparseMatrixReadPstart(): Pstart: Read Error") ;
    }

} /* end MMSparseMatrixReadPstart */

  
void MMSparseMatrixReadWeights(struct sparsematrix *pM, FILE *fp) {

    /* This function reads the row and column weights of a weighted
       sparse matrix. The weights are integers. 
       
       The total number of row weights is M.NrRowWeights.
       The total number of column weights is M.NrColWeights. 
       Each weight should be on a separate line. */

    long t ;
  
    if ( pM->MMTypeCode[0] == 'W' ){
  
        for ( t = 0 ; t < pM->NrRowWeights ; t++  )
            if ( fscanf(fp, "%ld\n", &(pM->RowWeights[t])) != 1 )
                Abort("MMSparseMatrixReadWeights(): Weights: Read Error") ;
  
        for ( t = 0 ; t < pM->NrColWeights ; t++  )
            if ( fscanf(fp, "%ld\n", &(pM->ColWeights[t])) != 1 )
                Abort("MMSparseMatrixReadWeights(): Weights: Read Error") ;
    }

} /* end MMSparseMatrixReadWeights */
  

void MMSparseMatrixReadEntries(struct sparsematrix *pM, FILE *fp) {

    /* This function reads the nz(M) nonzero entries of a sparse matrix M
       in Matrix Market format from the input stream fp.

       An entry is:
       - a pair (i, j) for a pattern matrix,
         where i is the row index and j the column index,
         with 1 <= i <= m and 1 <= j <= n.
         Matrix Market indices are numbered starting from 1.
         The indices are converted to internal indices of the sparse
         matrix data structure, which are numbered starting from 0.
       - a triple (i, j, val) for a real or integer matrix,
         where val is the numerical value. This value is a real,
         or it is converted to a real.
       - a quadruple (i, j, ReVal, ImVal) for a complex matrix,
         where ReVal is the real part of the numerical value and
         ImVal the imaginary part. 
         
      Each entry should be on a separate line. */

    long t ;
  
    if ( pM->MMTypeCode[2] == 'P' ) { /* pattern matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
            if ( fscanf(fp, "%ld %ld\n", &(pM->i[t]), &(pM->j[t])) != 2 )
                Abort("MMSparseMatrixReadEntries(): Pattern: Read Error") ;
            if (pM->i[t] < 1 || pM->i[t] > pM->m || 
                pM->j[t] < 1 || pM->j[t] > pM->n )
                Abort("MMSparseMatrixReadEntries(): Pattern: Read Error") ;
            pM->i[t]-- ; /* decrement by 1, since internally the
                            row and column indices of sparse matrices
                            are numbered from 0 */
            pM->j[t]-- ;
        }
    } else if ( pM->MMTypeCode[2] == 'I' ) { /* integer matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
            if ( fscanf(fp, "%ld %ld %lg\n", &(pM->i[t]), &(pM->j[t]), 
                 &(pM->ReValue[t])) != 3 ) 
                Abort("MMSparseMatrixReadEntries(): Integer: Read Error") ;
            if (pM->i[t] < 1 || pM->i[t] > pM->m ||  
                pM->j[t] < 1 || pM->j[t] > pM->n )
                Abort("MMSparseMatrixReadEntries(): Integer: Read Error") ;
            pM->i[t]-- ;
            pM->j[t]-- ;
        }
    } else if ( pM->MMTypeCode[2] == 'R' ) { /* real matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
            if ( fscanf(fp, "%ld %ld %lg\n", &(pM->i[t]), &(pM->j[t]), 
                        &(pM->ReValue[t])) != 3 )
                Abort("MMSparseMatrixReadEntries(): Real: Read Error") ; 
            if (pM->i[t] < 1 || pM->i[t] > pM->m ||  
                pM->j[t] < 1 || pM->j[t] > pM->n )
                Abort("MMSparseMatrixReadEntries(): Real: Read Error") ; 
            pM->i[t]-- ;
            pM->j[t]-- ;
        }
    } else if ( pM->MMTypeCode[2] == 'C' ) { /* complex matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
            if ( fscanf(fp, "%ld %ld %lg %lg\n", &(pM->i[t]), &(pM->j[t]), 
                        &(pM->ReValue[t]), &(pM->ImValue[t])) != 4 )
                Abort("MMSparseMatrixReadEntries(): Complex: Read Error") ;
            if (pM->i[t] < 1 || pM->i[t] > pM->m ||  
                pM->j[t] < 1 || pM->j[t] > pM->n )
                Abort("MMSparseMatrixReadEntries(): Complex: Read Error") ;
            pM->i[t]-- ;
            pM->j[t]-- ;
        }
    }
    
    /* Check if symmetric or hermitian matrix has entries
       only on or below the diagonal */
    if ( pM->MMTypeCode[3] == 'S' || pM->MMTypeCode[3] == 'H' )
        for ( t = 0 ; t < pM->NrNzElts ; t++ )
            if (pM->i[t] < pM->j[t] )
                Abort("MMSparseMatrixReadEntries(): Symmetric: Read Error") ;

     /* Check if skew-symmetric matrix has entries only below the diagonal */
    if ( pM->MMTypeCode[3] == 'K')
        for ( t = 0 ; t < pM->NrNzElts ; t++ )
            if (pM->i[t] <= pM->j[t] )
                Abort("MMSparseMatrixReadEntries(): Skew-symmetric: Read Error") ;
  
  } /* end MMSparseMatrixReadEntries */
  

void MMSparseMatrixReadTail(struct sparsematrix *pM, FILE *fp) {
    
    /* This function reads the tail comment from the input stream fp
       into the string M.tail. It removes initial white space characters
       (blanks, tabs) from each line, and empty lines  */
    
    int first, last ;
    char *line, linebuffer[MAX_LINE_LENGTH] ; 
    long tsize ;
  
    pM->tail = NULL ;
    tsize = 1 ; /* tail size (in characters) */
    while ( (line = fgets(linebuffer, MAX_LINE_LENGTH, fp)) != NULL ) { 
        /* a new line has been read */
        first = 0 ; 
        last = strlen(line) ;
        if ( linebuffer[last] != '\0' )
            Abort("MMSparseMatrixReadTail(): internal error") ;
  
        /* Remove initial blank characters from the line */
        while ( isspace((int) linebuffer[first]) ) 
            first++ ;
  
        line = linebuffer + first ;
        
        if ( first != last ) {
            /* Add this line to the tail */
            tsize += last - first ;
            pM->tail = (char *) realloc(pM->tail, tsize * sizeof(char)) ; 
            if ( pM->tail == NULL )
                Abort("MMSparseMatrixReadTail(): Not enough memory for realloc "
                      "of matrix tail") ;
            pM->tail[tsize + first - last - 1] = '\0' ;
            strncat(pM->tail, line, last-first) ;
        }
    }
       
} /* end MMSparseMatrixReadTail */
  
  
void MMSparseMatrixPrintHeader(struct sparsematrix *pM, FILE *stream) {

    /* This function prints the header of a sparse matrix M
       to the output stream.

       The header is a banner line with the types of the matrix,
       followed by the comment lines as stored in M.header,
       followed by a line with the size of the matrix,
           m n nz
       where m = the number of rows, m >=1
             n = the number of columns, n >= 1
             nz the number of nonzeros, nz >= 1.
       
       For a distributed matrix, the size line has an extra parameter:
           m n nz P
       where P is the number of processors, P >= 1.

       For a weighted matrix, the size line is
           m n nz param
       where param = 0 means no row or column weights
                     1       only row weights
                     2       only column weights
                     3       both row and column weights.
    */
       
    char line[MAX_LINE_LENGTH] ; 
    int param ;
  
    /* Print Matrix Market type codes to first line */
    MMSparseMatrixGetTypeCode(pM, line) ;
    fprintf(stream, "%s", line) ;
  
    /* Print the header */
    if ( pM->header != NULL )
        fprintf(stream, "%s", pM->header) ;
  
    /* Print m, n, nz, and perhaps an extra parameter to the first line */
    if ( pM->MMTypeCode[0] == 'D' ) /* distributed matrix */
        fprintf(stream, "%ld %ld %ld %d\n", pM->m, pM->n, pM->NrNzElts, 
                pM->NrProcs) ;
    else if ( pM->MMTypeCode[0] == 'W' ){ /* weighted matrix */
        /* Compute parameter */
        param = 0 ;
        if (pM->NrRowWeights > 0)
            param++ ;
        if (pM->NrColWeights > 0)
            param += 2;
        fprintf(stream, "%ld %ld %ld %d\n", pM->m, pM->n, pM->NrNzElts,
                param) ; 
    } else
        fprintf(stream, "%ld %ld %ld\n", pM->m, pM->n, pM->NrNzElts) ;

} /* end MMSparseMatrixPrintHeader */
  

void MMSparseMatrixPrintPstart(struct sparsematrix *pM, FILE *stream) {

    /* This function prints the start of the nonzeros for each of the
       P processor parts, where P = M.NrProcs. The nonzeros of processor q
       start at position Pstart[q] and end at Pstart[q+1]-1, for 0 <= q < P.
       For convenience, Pstart[0] = 0 and Pstart[P] = nz(M),
       but the function will print the values whatever they are.
       Each Pstart[q] value is printed on a separate line. */

    int t ;
  
    if ( pM->MMTypeCode[0] == 'D' ) 
        for ( t = 0 ; t <= pM->NrProcs ; t++  )
            fprintf(stream, "%ld\n", pM->Pstart[t]) ;

} /* end MMSparseMatrixPrintPstart */

  
void MMSparseMatrixPrintWeights(struct sparsematrix *pM, FILE *stream) {

    /* This function prints the row and column weights of a weighted
       sparse matrix. The weights are integers.

       The total number of row weights is M.NrRowWeights.
       The total number of column weights is M.NrColWeights.
       Each weight is printed on a separate line. */

    long t ;
  
    if ( pM->MMTypeCode[0] == 'W' ){
        for ( t = 0 ; t < pM->NrRowWeights ; t++  )
            fprintf(stream, "%ld\n", pM->RowWeights[t]) ;
        for ( t = 0 ; t < pM->NrColWeights ; t++  )
            fprintf(stream, "%ld\n", pM->ColWeights[t]) ;
    }

} /* end MMSparseMatrixPrintWeights */
  
  
void MMSparseMatrixPrintEntries(struct sparsematrix *pM, FILE *stream) {

    /* This function prints the nz(M) nonzero entries of a sparse matrix M
       in Matrix Market format to the output stream.

       An entry is:
       - a pair (i, j) for a pattern matrix,
         where i is the row index and j the column index,
         with 1 <= i <= m and 1 <= j <= n.
         Matrix Market indices are numbered starting from 1.
         The indices are converted from internal indices of the sparse
         matrix data structure, which are numbered starting from 0.
       - a triple (i, j, val) for a real or integer matrix,
         where val is the numerical value.
       - a quadruple (i, j, ReVal, ImVal) for a complex matrix,
         where ReVal is the real part of the numerical value and
         ImVal the imaginary part.

      Each entry will be printed on a separate line. */

    long t, r ;
  
    if ( pM->MMTypeCode[2] == 'P' ) { /* pattern matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ )
            fprintf(stream, "%ld %ld\n", pM->i[t]+1, pM->j[t]+1) ;
    } else if ( pM->MMTypeCode[2] == 'I' ) { /* integer matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
            /* round to the nearest integer */
            if (pM->ReValue[t] >= 0)
                r = pM->ReValue[t]+0.5 ; 
            else 
                r = pM->ReValue[t]-0.5 ;
            fprintf(stream, "%ld %ld %ld\n", pM->i[t]+1, pM->j[t]+1, r) ;
        }
    } else if ( pM->MMTypeCode[2] == 'R' ) { /* real matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ )
            fprintf(stream, "%ld %ld %g\n", pM->i[t]+1, pM->j[t]+1, pM->ReValue[t]) ;
    } else if ( pM->MMTypeCode[2] == 'C' ) { /* complex matrix */
        for ( t = 0 ; t < pM->NrNzElts ; t++ )
            fprintf(stream, "%ld %ld %g %g\n", pM->i[t]+1, pM->j[t]+1, 
                    pM->ReValue[t], pM->ImValue[t]) ;
    }

} /* end MMSparseMatrixPrintEntries */

  
void MMSparseMatrixPrintTail(struct sparsematrix *pM, FILE *stream) {

    /* This function prints the tail comment of a square matrix 
       to the output stream */ 

    if ( pM->tail != NULL )
        fprintf(stream, "%s", pM->tail) ; 
        
} /* end MMSparseMatrixPrintTail */
  
  
void AddDummiesToSparseMatrix(struct sparsematrix *pM)  {
  
    /* This function adds dummies to a square matrix to make its diagonal 
       structurally nonzero. It also creates a boolean array M.dummy
       such that M.dummy[i] = TRUE if entry (i,i) is a dummy nonzero.
       
       A dummy is an entry in the sparse matrix data structure with
       numerical value 0.
       
       This function should only be used before matrix partitioning,
       because it does not update distribution information. */
  
    long t, tt ;
    long i, j, n, NrNzEltsNew ;
  
    if ( pM->m != pM->n )
        Abort("AddDummiesToSparseMatrix(): Matrix is not square !") ;
  
    n = pM->n ;
    pM->NrDummies = n ; /* maximum number of dummies needed */
    
    /* Allocate memory for dummy information */
    pM->dummy = (int *) malloc(n * sizeof(int)) ;
    if ( pM->dummy == NULL )
        Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies") ;
  
    /* Initialise dummy information */
    for ( i = 0 ; i < n ; i++ )
        pM->dummy[i] = TRUE ; /* dummy needed for matrix element (i,i) */
  
    /* Fill dummy and count the number of dummies */
    for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
        i = pM->i[t] ;
        j = pM->j[t] ;
        if ( i == j )
            if ( pM->dummy[i] ) {
                pM->dummy[i] = FALSE ;
                pM->NrDummies-- ; /* no dummy needed because of actual
                                     diagonal nonzero */
            }
    }
  
    /* Allocate memory for dummies. Allocate one element extra
       in case the matrix becomes empty. */
    NrNzEltsNew = pM->NrNzElts + pM->NrDummies ; 

    pM->i = (long *) realloc(pM->i, (NrNzEltsNew+1) * sizeof(long)) ;
    pM->j = (long *) realloc(pM->j, (NrNzEltsNew+1) * sizeof(long)) ;
    if ( pM->i == NULL || pM->j == NULL )
        Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies");
    if ( pM->MMTypeCode[2] != 'P' ) {
        pM->ReValue = (double *) realloc(pM->ReValue,
                                         (NrNzEltsNew+1) * sizeof(double)) ;
        if ( pM->ReValue == NULL )
            Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies");
    }
    if ( pM->MMTypeCode[2] == 'C' ) {
        pM->ImValue = (double *) realloc(pM->ImValue,
                                         (NrNzEltsNew+1) * sizeof(double)) ;
        if ( pM->ImValue == NULL )
            Abort("AddDummiesToSparseMatrix(): Not enough memory for dummies");
    }
  
    /* Add the dummies to the matrix */
    tt = pM->NrNzElts ;
    for ( i = 0 ; i < n ; i++) {
        if ( pM->dummy[i] ) {
            pM->i[tt] = i ;
            pM->j[tt] = i ;
            if ( pM->MMTypeCode[2] != 'P' )
                pM->ReValue[tt] = 0.0 ;
            if ( pM->MMTypeCode[2] == 'C' )
                pM->ImValue[tt] = 0.0 ; 
            tt++ ;
        }
    }
  
    /* Update the number of nonempty elements of the matrix */
    pM->NrNzElts = NrNzEltsNew ;

} /* end AddDummiesToSparseMatrix */
  
  
void RemoveDummiesFromSparseMatrix(struct sparsematrix *pM)  {
  
    /* This function removes all dummies from a square matrix
       and frees the memory of the boolean array M.dummy.

       A dummy is an entry in the sparse matrix data structure with
       numerical value 0.

       This function can be used before and after partitioning,
       because it updates the distribution information by adjusting Pstart. */
  
    long i, j, t, tt, n ;
    int q ;
    long NrRemoved, start ;
    
    /* If the matrix contains no dummies, free M.dummy and return */
    if ( pM->NrDummies <= 0 ) {
        if ( pM->dummy != NULL )
            free(pM->dummy) ;
        return ;
    }
  
    /* If the matrix is not square, there should be no dummies ! */
    if ( pM->m != pM->n )
        Abort("RemoveDummiesFromSparseMatrix(): Matrix is not square") ;
    n =  pM->n ;
  
    if ( pM->NrProcs < 1 || pM->Pstart == NULL ) { /* matrix not partitioned */
        NrRemoved = 0 ;
        t = 0 ;
        while ( t < pM->NrNzElts ) {
            while ( (i = pM->i[t]) == pM->j[t] && NrRemoved < pM->NrDummies &&
                     pM->dummy[i] ) { 
                tt = pM->NrNzElts-1 ;
    
                /* Swap dummy t with last entry tt */
                SwapLong(pM->i, t, tt) ; /* pM->i[t] <-> pM->i[tt] */
                SwapLong(pM->j, t, tt) ;
                if ( pM->MMTypeCode[2] != 'P' ) 
                    SwapDouble(pM->ReValue, t, tt) ; 
                if ( pM->MMTypeCode[2] == 'C' ) 
                    SwapDouble(pM->ImValue, t, tt) ;
      
                /* Remove last element */
                pM->NrNzElts-- ;
                NrRemoved++ ;

                /* entry t could again be a dummy, hence we need a loop */
            }
            t++ ;
        }
    } else { /* Update Pstart as well */
        NrRemoved = 0 ;
        start = pM->Pstart[0] ;
        tt = 0 ;
        for ( q = 0 ; q < pM->NrProcs ; q++ ) {
            for ( t = start ; t < pM->Pstart[q+1] ; t++ ) {
                i = pM->i[t] ;
                j = pM->j[t] ;
                if (i == j && pM->dummy[i] )
                    NrRemoved++ ; /* remove dummy */
                else { /* Move nondummy to front position tt */
                    pM->i[tt] = i ;
                    pM->j[tt] = j;
                    if ( pM->MMTypeCode[2] != 'P' )
                        pM->ReValue[tt] = pM->ReValue[t] ;
                    if ( pM->MMTypeCode[2] == 'C' )
                        pM->ImValue[tt] = pM->ImValue[t] ;
                    tt++ ;
                }
            }
    
            /* Update Pstart */
            start = pM->Pstart[q+1] ;
            pM->Pstart[q+1] = tt ;
        }
  
        /* Update NrNzElts */
        pM->NrNzElts -= NrRemoved ;
    }
  
    /* Check the number of removed dummies */
    if ( pM->NrDummies != NrRemoved ) 
        Abort("RemoveDummiesFromSparseMatrix(): Internal error: "
              "NrDummies != NrRemoved ") ;
  
    /* Free the memory of the dummies */
    pM->i = (long *) realloc(pM->i, (pM->NrNzElts+1) * sizeof(long)) ;
    pM->j = (long *) realloc(pM->j, (pM->NrNzElts+1) * sizeof(long)) ;
    if ( pM->i == NULL || pM->j == NULL )
        Abort("RemoveDummiesFromSparseMatrix(): Not enough memory") ;
    if ( pM->MMTypeCode[2] != 'P' ) {
        pM->ReValue = (double *) realloc(pM->ReValue, (pM->NrNzElts+1) * 
                                         sizeof(double)) ;
        if ( pM->ReValue == NULL )
            Abort("RemoveDummiesFromSparseMatrix(): Not enough memory") ;
    }
    if ( pM->MMTypeCode[2] == 'C' ) {
        pM->ImValue = (double *) realloc(pM->ImValue, (pM->NrNzElts+1) * 
                                         sizeof(double)) ;
        if ( pM->ImValue == NULL )
            Abort("RemoveDummiesFromSparseMatrix(): Not enough memory") ;
    } 
  
    if ( pM->dummy != NULL )
        free(pM->dummy) ;
    pM->NrDummies = 0 ;

} /* end RemoveDummiesFromSparseMatrix */

  
void SparseMatrixSymmetric2Full(struct sparsematrix *pM) {

    /* This function converts symmetric, skew-symmetric,
       and hermitian matrices to the full sparse representation.

       A symmetric matrix (a[i,j] = a[j,i]) or hermitian matrix
       (a[i,j] = conjg(a[j,i])) stores only the nonzeros on or 
       below the diagonal.

       A skew-symmetric (or anti-symmetric) matrix (a[i,j] = -a[j,i])
       stores only the nonzeros below the diagonal.

       This representation is also called the lower triangular format.
       In the full sparse representation all nonzeros are stored, 
       including those above the diagonal. 

       This function can be used before and after partitioning.
       If it is called after partitioning, it assumes
       symmetric partitioning by assigning a[i,j] and a[j,i]
       to the same processor.

    */
  
    long t, tt, Ndiag ;
    long start, NrNzEltsNew ;
    int q ;
  
    /* Check if matrix is symmetric */
    if ( pM->m != pM->n || 
         (pM->MMTypeCode[3] != 'S' && pM->MMTypeCode[3] != 'K' &&
          pM->MMTypeCode[3] != 'H' ))
        Abort("SparseMatrixSymmetric2Full(): matrix is not symmetric") ;
  
    /* Count the number of diagonal elements */
    Ndiag = 0 ;
    for ( t = 0 ; t < pM->NrNzElts ; t++ )
        if (pM->i[t] == pM->j[t])
            Ndiag++ ;
    NrNzEltsNew = 2 * pM->NrNzElts - Ndiag;

    if ( pM->MMTypeCode[3] == 'K' && Ndiag > 0 )
        Abort("SparseMatrixSymmetric2Full(): matrix is not skew-symmetric") ;

    /* Update Pstart */
    if ( pM->NrProcs >= 1 && pM->Pstart != NULL ) {
        tt = 0 ;
        start = pM->Pstart[0] ;
        for ( q = 0 ; q < pM->NrProcs ; q++ ) {
            for ( t = start ; t < pM->Pstart[q+1] ; t++ ) {
                if ( pM->i[t] == pM->j[t] )
                    tt++ ;
                else
                    tt += 2;
            }    
           
            start = pM->Pstart[q+1] ;
            pM->Pstart[q+1] = tt ;
        }
    }
   
    /* Allocate memory for the entries */
    pM->i = (long *) realloc(pM->i, (NrNzEltsNew+1) * sizeof(long)) ;
    pM->j = (long *) realloc(pM->j, (NrNzEltsNew+1) * sizeof(long)) ;
    if ( pM->i == NULL || pM->j == NULL )
        Abort("SparseMatrixSymmetric2Full(): Not enough memory to store"
              "full matrix") ;
    if ( pM->MMTypeCode[2] != 'P' ) {
        pM->ReValue = (double *) realloc(pM->ReValue, (NrNzEltsNew+1) * 
                                         sizeof(double)) ;
        if ( pM->ReValue == NULL )
            Abort("SparseMatrixSymmetric2Full(): Not enough memory to store"
                  "full matrix") ;
    }
    if ( pM->MMTypeCode[2] == 'C' ) {
        pM->ImValue = (double *) realloc(pM->ImValue, (NrNzEltsNew+1) * 
                                         sizeof(double)) ;
        if ( pM->ImValue == NULL )
            Abort("SparseMatrixSymmetric2Full(): Not enough memory to store"
                  "full matrix") ;
    }
  
    /* Copy the entries */
    tt = NrNzEltsNew -1 ;
    for ( t = pM->NrNzElts - 1 ; t >= 0; t-- ) {
        /* Copy entry t into tt */
        if ( pM->i[t] <  pM->j[t])
            Abort("SparseMatrixSymmetric2Full(): entry above diagonal") ;

        pM->i[tt] = pM->i[t] ;
        pM->j[tt] = pM->j[t] ;
        if ( pM->MMTypeCode[2] != 'P' )
            pM->ReValue[tt] = pM->ReValue[t] ;
        if ( pM->MMTypeCode[2] == 'C' )
            pM->ImValue[tt] = pM->ImValue[t] ;
        tt-- ;   
  
        /* Copy and transpose nondiagonal entry t */
        if (pM->i[tt+1] != pM->j[tt+1]) {
            pM->i[tt] = pM->j[tt+1] ; 
            pM->j[tt] = pM->i[tt+1] ; 
            if ( pM->MMTypeCode[2] != 'P' ) {
                if ( pM->MMTypeCode[3] == 'K' )
                    pM->ReValue[tt] = -pM->ReValue[tt+1] ; 
                else
                    pM->ReValue[tt] = pM->ReValue[tt+1] ; 
            }
            if ( pM->MMTypeCode[2] == 'C' ) { 
                if ( pM->MMTypeCode[3] == 'K' || pM->MMTypeCode[3] == 'H' ) 
                    pM->ImValue[tt] = -pM->ImValue[tt+1] ;
                else 
                    pM->ImValue[tt] = pM->ImValue[tt+1] ;
            } 
            tt-- ;
        }
    }
  
    /* Update the number of nonzero entries and the matrix type code */
    pM->NrNzElts = NrNzEltsNew;
    pM->MMTypeCode[3] = 'G' ;
    strcpy(pM->Symmetry, "general") ;
   
} /* end SparseMatrixSymmetric2Full */
  

void SparseMatrixSymmetricLower2Random(struct sparsematrix *pM) {

    /* This function converts a symmetric matrix stored
       in lower triangular format (with only nonzeros a[i,j] with
       i <= j present) into a symmetric matrix stored in random format,
       where the numerical value a[i,j] with i <= j is assigned
       either to entry (i,j) or to (j,i) in the sparse matrix
       data structure, each with equal probability.

       This conversion is done by transposing the indices of every nonzero
       with a probability of 50 percent, without changing the numerical value.
       This function preserves the distribution information.
    */

    long t, tmp ;
  
    /* Check if matrix is symmetric */
    if ( pM->m != pM->n || 
         (pM->MMTypeCode[3] != 'S'  && pM->MMTypeCode[3] != 'K'  &&
          pM->MMTypeCode[3] != 'H' ) ) 
        Abort("SparseMatrixSymmetricLower2Random(): matrix is not symmetric") ;
    
    for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
        if (Random1(0,1) == 0) {
            /* Swap index i and j */
            tmp = pM->i[t] ;
            pM->i[t] = pM->j[t] ;
            pM->j[t] = tmp ;
        }
    }

} /* end SparseMatrixSymmetricLower2Random */

  
void SparseMatrixSymmetricRandom2Lower(struct sparsematrix *pM) {

    /* This function converts a symmetric matrix stored
       in random format, where the numerical value a[i,j] with i <= j
       is assigned either to entry (i,j) or to (j,i) in the sparse matrix
       data structure, not necesssarily with equal probability,
       into a symmetric matrix stored in lower triangular format
       (with only nonzeros a[i,j] with i <= j present).

       This conversion is done by transposing the indices of every nonzero
       (i,j) with i < j in the data structure.

       This function preserves the distribution information.
    */


    long t, tmp ;
  
    /* Check if matrix is symmetric */
    if ( pM->m != pM->n || 
         (pM->MMTypeCode[3] != 'S'  && pM->MMTypeCode[3] != 'K'  &&
          pM->MMTypeCode[3] != 'H' ) )   
        Abort("SparseMatrixSymmetricRandom2Lower(): matrix is not symmetric") ;
   
    for ( t = 0 ; t < pM->NrNzElts ; t++ ) {
        if ( pM->i[t] < pM->j[t] ) {
            /* Swap index i and j */
            tmp = pM->i[t] ;
            pM->i[t] = pM->j[t] ;
            pM->j[t] = tmp ;
        }
    }

} /* end SparseMatrixSymmetricRandom2Lower */


void SparseMatrixRemoveDuplicates(struct sparsematrix *pM) {

    /* This function removes the duplicate nonzeros from a sparse matrix
       by adding their numerical values, giving a single value at a location (i,j).

       This function is intended for checking the input of a sparse matrix after it has 
       been read from a file, and repairing it if necessary. 

       This function can only be used before partitioning and before adding dummies,
       as it does not preserve the distribution information nor the dummy information.
    */

    long nz, t, tt, k, kpred, *K;
    
    nz = pM->NrNzElts ;
    
    /* Allocate memory for nonzero index array */   
    K = (long *) malloc(nz * sizeof(long)) ;
    if ( K == NULL )
        Abort("SparseMatrixRemoveDuplicates(): Not enough memory") ;
    for ( t = 0 ; t < nz ; t++ )
        K[t] = t ;
        
    /* Sort nonzero indices stably, first by column index, then by row index.
       As a result, the indices K[t] give the order of the nonzeros
       by increasing row index, and within the same row by increasing column index */

    CSort(K, pM->j, pM->n-1, 0, nz-1) ;
    CSort(K, pM->i, pM->m-1, 0, nz-1) ;

    /* Check if a nonzero is at the same location as its predecessor.
       If so, add the numerical values and mark the predecessor for removal */
    for ( t = 1 ; t < nz ; t++ ) {
        k = K[t] ;
        kpred = K[t-1] ;
        if (pM->i[k] == pM->i[kpred] && pM->j[k] == pM->j[kpred] ) {
            /* add the numerical values */
            if ( pM->MMTypeCode[2] != 'P' ) 
                pM->ReValue[k] +=  pM->ReValue[kpred] ;
            if ( pM->MMTypeCode[2] == 'C' )
                pM->ImValue[k] +=  pM->ImValue[kpred] ;
        
            /* mark the predecessor for removal */
            pM->i[kpred] = -1 ;
        }
    }
    
    /* Remove the marked nonzeros */
    tt = 0 ;
    for ( t = 0 ; t < nz ; t++ ) {
        if (pM->i[t] != -1){
            /* copy nonzero from position t into tt */
            pM->i[tt] = pM->i[t];
            pM->j[tt] = pM->j[t]; 
            if ( pM->MMTypeCode[2] != 'P' )
                pM->ReValue[tt] = pM->ReValue[t] ;
            if ( pM->MMTypeCode[2] == 'C' )
                pM->ImValue[tt] = pM->ImValue[t] ;
            tt++ ;
        }
    }
    
#ifdef INFO
    printf("  %ld duplicate nonzeros removed\n", nz-tt) ;
    printf("  New number of nonzeros: %ld \n", tt) ;
#endif

    pM->NrNzElts = tt ;
    
    free(K) ; 
    
} /* end SparseMatrixRemoveDuplicates */


int SparseMatrixStructurallySymmetric(struct sparsematrix *pM) {

    /* This function checks whether a sparse matrix is 
       structurally symmetric, i.e, it is square and 
       A[i,j] is a nonzero (in the data structure) if and only if A[j,i] is a nonzero,
       for all i,j. If this holds, the value TRUE is returned; otherwise, FALSE. 
       
       The numerical value of the nonzero A[i,j] does not matter.
       It can even be 0.0 (an accidental zero).     

       This function does not change the matrix.
    */

    long nz, t, kr, kc, *Kr, *Kc ;
    int symmetric ;
    
    /* Check whether the matrix is square */
    if (pM->m != pM->n)
        return (int)(FALSE) ;

    /* Check if matrix is structurally symmetric by definition */
    if ( (pM->MMTypeCode[3] == 'S' || pM->MMTypeCode[3] == 'K' ||
          pM->MMTypeCode[3] == 'H' ))
              return (int)(TRUE) ;
        
    nz = pM->NrNzElts ;
   
    /* Allocate memory for nonzero index arrays Kr and Kc */   
    Kr = (long *) malloc(nz * sizeof(long)) ;
    Kc = (long *) malloc(nz * sizeof(long)) ;
    if ( Kr == NULL || Kc == NULL )
        Abort("SparseMatrixStructurallySymmetric(): Not enough memory") ;
    for ( t = 0 ; t < nz ; t++ ) {
        Kr[t] = t ;        
        Kc[t] = t ;
    }

    /* Sort nonzero indices of Kr stably, first by column index, then by row index.
       As a result, the indices Kr[t] give the order of the nonzeros
       by increasing row index, and within the same row by increasing column index */
 
    CSort(Kr, pM->j, pM->n-1, 0, nz-1) ;
    CSort(Kr, pM->i, pM->n-1, 0, nz-1) ;
    
    /* Sort nonzero indices of Kc stably, first by row index, then by column index.
       As a result, the indices Kc[t] give the order
       by increasing column index, and within the same column by increasing row index */

    CSort(Kc, pM->i, pM->n-1, 0, nz-1) ;
    CSort(Kc, pM->j, pM->n-1, 0, nz-1) ;

    /* Check if each nonzero (i,j) also has a nonzero in position (j,i) */
    symmetric = TRUE ;
    for ( t = 0 ; t < nz ; t++ ) {
        kr = Kr[t] ;
        kc = Kc[t] ;
        if (pM->i[kr] != pM->j[kc] || pM->j[kr] != pM->i[kc] ) {
            symmetric = FALSE ;
            break ;
        }
    }
    
    free(Kc) ; 
    free(Kr) ; 
    
    return symmetric ;
    
} /* end  SparseMatrixStructurallySymmetric */

