#ifndef __SparseMatrix_h__
#define __SparseMatrix_h__

#include "IO.h"
#include "Sort.h"

#define MM_Banner           "%%MatrixMarket"

/* Sparse matrix data structure containing a list of triples (i,j,a)
   representing nonzeros, where i is the row index, j the column index,
   and a the numerical value */
    
typedef struct sparsematrix {
  char MMTypeCode[4] ; 
  char Banner[MAX_WORD_LENGTH] ;
  char Object[MAX_WORD_LENGTH] ;
  char Format[MAX_WORD_LENGTH] ;
  char Field[MAX_WORD_LENGTH] ;
  char Symmetry[MAX_WORD_LENGTH] ;

  /* Matrix dimensions */
  long NrNzElts ;  /* Number of nonzero matrix elements */
  long m ;         /* Number of matrix rows */
  long n ;         /* Number of matrix columns */
  long NrDummies ; /* Number of dummy diagonal nonzeros */
  int  NrProcs ;   /* Number of processors for a distributed matrix */

  /* Arrays of length NrNzElts storing the triples */
  long *i ;         /* List of row indices */
  long *j ;         /* List of column indices */
  double *ReValue ; /* List of Real part of the matrix values */
  double *ImValue ; /* List of Imaginary part of the matrix values */
  
  /* Additional arrays */
  int *dummy ;   /* Boolean array which says whether a diagonal element
                   is dummy or not */
  long *Pstart ; /* Pointer to the start of the triples for the processors */

  char *header ; /* Header comment of the matrix */
  char *tail ;   /* Tail comment of the matrix */

  long NrRowWeights ; /* Number of row weights, = 0 or m */
  long NrColWeights ; /* Number of column weights, = 0 or n */
  long *RowWeights ;  /* Array of row weights */
  long *ColWeights ;  /* Array of column weights */

  int *RowMark ; /* Array for marking rows, can be used e.g. to check
                    whether a row is cut by partitioning */
  int *ColMark ;

} SparseMatrix ;

/* Function declarations for Matrix Market input/output functions */
void MMSparseMatrixInit(struct sparsematrix *pM) ;
struct sparsematrix MMReadSparseMatrix(char *filename) ;
void MMWriteSparseMatrix(struct sparsematrix M, char *filename) ;
void MMDeleteSparseMatrix(struct sparsematrix *pM) ;
void MMSparseMatrixSetTypeCode(struct sparsematrix *pM, char *line) ;
void MMSparseMatrixGetTypeCode(struct sparsematrix *pM, char *line) ;
void MMSparseMatrixAllocateMemory(struct sparsematrix *pM) ;
void MMSparseMatrixFreeMemory(struct sparsematrix *pM) ;

void MMSparseMatrixReadHeader(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadPstart(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadWeights(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadEntries(struct sparsematrix *pM, FILE *fp) ;
void MMSparseMatrixReadTail(struct sparsematrix *pM, FILE *fp) ;

void MMSparseMatrixPrintHeader(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintPstart(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintWeights(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintEntries(struct sparsematrix *pM, FILE *stream) ;
void MMSparseMatrixPrintTail(struct sparsematrix *pM, FILE *stream) ;

/* Function declarations for matrix conversion functions */
void AddDummiesToSparseMatrix(struct sparsematrix *pM) ;
void RemoveDummiesFromSparseMatrix(struct sparsematrix *pM) ;

void SparseMatrixSymmetric2Full(struct sparsematrix *pM) ;
void SparseMatrixSymmetricLower2Random(struct sparsematrix *pM) ;
void SparseMatrixSymmetricRandom2Lower(struct sparsematrix *pM) ;

void SparseMatrixRemoveDuplicates(struct sparsematrix *pM) ;
int SparseMatrixStructurallySymmetric(struct sparsematrix *pM) ;

#endif /* __SparseMatrix_h__ */
